'''
交易大数据
https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api
'''
from paux.param import to_local
from okx.api._client import Client


class _TradingStatisticsEndpoints():
    get_support_coin = ['/api/v5/rubik/stat/trading-data/support-coin', 'GET']  # 获取交易大数据支持币种
    get_taker_volume = ['/api/v5/rubik/stat/taker-volume', 'GET']  # 获取主动买入/卖出情况
    get_loan_ratio = ['/api/v5/rubik/stat/margin/loan-ratio', 'GET']  # 获取杠杆多空比
    get_long_short_account_ratio = ['/api/v5/rubik/stat/contracts/long-short-account-ratio', 'GET']  # 获取合约多空持仓人数比
    get_contracts_open_interest_volume = ['/api/v5/rubik/stat/contracts/open-interest-volume', 'GET']  # 获取合约持仓量及交易量
    get_option_open_interest_volume = ['/api/v5/rubik/stat/option/open-interest-volume', 'GET']  # 获取期权持仓量及交易量
    get_open_interest_volume_ratio = ['/api/v5/rubik/stat/option/open-interest-volume-ratio', 'GET']  # 看涨/看跌期权合约 持仓总量比/交易总量比
    get_open_interest_volume_expiry = ['/api/v5/rubik/stat/option/open-interest-volume-expiry', 'GET']  # 看涨看跌持仓总量及交易总量（按到期日分）
    get_open_interest_volume_strike = ['/api/v5/rubik/stat/option/open-interest-volume-strike', 'GET']  # 看涨看跌持仓总量及交易总量（按执行价格分）
    get_taker_block_volume = ['/api/v5/rubik/stat/option/taker-block-volume', 'GET']  # 看跌/看涨期权合约 主动买入/卖出量


class TradingStatistics(Client):

    # 获取交易大数据支持币种
    def get_support_coin(self, proxies={}, proxy_host: str = None):
        '''
        获取支持交易大数据的币种
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-support-coin
        
        限速：5次/2s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        contract          	Array of strings	合约交易大数据接口功能支持的币种
        option            	Array of strings	期权交易大数据接口功能支持的币种
        spot              	Array of strings	现货交易大数据接口功能支持的币种
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_support_coin, **to_local(locals()))

    # 获取主动买入/卖出情况
    def get_taker_volume(self, ccy: str, instType: str, begin: str = '', end: str = '', period: str = '', proxies={},
                         proxy_host: str = None):
        '''
        获取taker主动买入和卖出的交易量
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-taker-volume
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        instType          	String  	是       	产品类型，币币："SPOT" , 衍生品："CONTRACTS"
        begin             	String  	否       	开始时间，例如：1597026383085
        end               	String  	否       	结束时间，例如：1597026383011
        period            	String  	否       	时间粒度，默认值5m。支持[5m/1H/1D]5m粒度最多只能查询两天之内的数据1H粒度最多只能查询30天之内的数据1D粒度最多只能查询180天之内的数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        sellVol           	String  	卖出量
        buyVol            	String  	买入量
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_taker_volume, **to_local(locals()))

    # 获取杠杆多空比
    def get_loan_ratio(self, ccy: str, begin: str = '', end: str = '', period: str = '', proxies={},
                       proxy_host: str = None):
        '''
        获取借入计价货币与借入交易货币的累计数额比值。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-margin-lending-ratio
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        begin             	String  	否       	开始时间，例如：1597026383085
        end               	String  	否       	结束时间，例如：1597026383011
        period            	String  	否       	时间粒度，默认值5m。支持[5m/1H/1D]5m粒度最多只能查询两天之内的数据1H粒度最多只能查询30天之内的数据1D粒度最多只能查询180天之内的数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        ratio             	String  	多空比值
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_loan_ratio, **to_local(locals()))

    # 获取合约多空持仓人数比
    def get_long_short_account_ratio(self, ccy: str, begin: str = '', end: str = '', period: str = '', proxies={},
                                     proxy_host: str = None):
        '''
        获取交割永续净开多持仓用户数与净开空持仓用户数的比值。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-long-short-ratio
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        begin             	String  	否       	开始时间，例如：1597026383085
        end               	String  	否       	结束时间，例如：1597026383011
        period            	String  	否       	时间粒度，默认值5m。支持[5m/1H/1D]5m粒度最多只能查询两天之内的数据1H粒度最多只能查询30天之内的数据1D粒度最多只能查询180天之内的数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        ratio             	String  	多空人数比
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_long_short_account_ratio, **to_local(locals()))

    # 获取合约持仓量及交易量
    def get_contracts_open_interest_volume(self, ccy: str, begin: str = '', end: str = '', period: str = '', proxies={},
                                           proxy_host: str = None):
        '''
        获取交割永续的持仓量和交易量。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-contracts-open-interest-and-volume
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        begin             	String  	否       	开始时间，例如：1597026383085
        end               	String  	否       	结束时间，例如：1597026383011
        period            	String  	否       	时间粒度，默认值5m。支持[5m/1H/1D]5m粒度最多只能查询两天之内的数据1H粒度最多只能查询30天之内的数据1D粒度最多只能查询180天之内的数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        oi                	String  	持仓总量（USD）
        vol               	String  	交易总量（USD）
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_contracts_open_interest_volume, **to_local(locals()))

    # 获取期权持仓量及交易量
    def get_option_open_interest_volume(self, ccy: str, period: str = '', proxies={}, proxy_host: str = None):
        '''
        获取期权的持仓量和交易量。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-options-open-interest-and-volume
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        period            	String  	否       	时间粒度，默认值8H。支持[8H/1D]每个粒度最多只能查询72条数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        oi                	String  	持仓总量（BTC）
        vol               	String  	交易总量（BTC）
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_option_open_interest_volume, **to_local(locals()))

    # 看涨/看跌期权合约 持仓总量比/交易总量比
    def get_open_interest_volume_ratio(self, ccy: str, period: str = '', proxies={}, proxy_host: str = None):
        '''
        获取看涨期权和看跌期权的持仓量比值，以及交易量比值。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-put-call-ratio
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        period            	String  	否       	时间粒度，默认值8H。支持[8H/1D]每个粒度最多只能查询72条数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        oiRatio           	String  	看涨/看跌 持仓总量比
        volRatio          	String  	看涨/看跌 交易总量比
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_open_interest_volume_ratio, **to_local(locals()))

    # 看涨看跌持仓总量及交易总量（按到期日分）
    def get_open_interest_volume_expiry(self, ccy: str, period: str = '', proxies={}, proxy_host: str = None):
        '''
        获取每个到期日上看涨期权和看跌期权的持仓量和交易量。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-open-interest-and-volume-expiry
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        period            	String  	否       	时间粒度，默认值8H。支持[8H/1D]每个粒度仅展示最新的一份数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        expTime           	String  	到期日（格式: YYYYMMDD，例如："20210623"）
        callOI            	String  	看涨持仓总量（以币为单位）
        putOI             	String  	看跌持仓总量（以币为单位）
        callVol           	String  	看涨交易总量（以币为单位）
        putVol            	String  	看跌交易总量（以币为单位）
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_open_interest_volume_expiry, **to_local(locals()))

    # 看涨看跌持仓总量及交易总量（按执行价格分）
    def get_open_interest_volume_strike(self, ccy: str, expTime: str, period: str = '', proxies={},
                                        proxy_host: str = None):
        '''
        获取看涨期权和看跌期权的taker主动买入和卖出的交易量。
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-open-interest-and-volume-strike
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        expTime           	String  	是       	到期日（格式: YYYYMMdd，例如："20210623"）
        period            	String  	否       	时间粒度，默认值8H。支持[8H/1D]每个粒度仅展示最新的一份数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        strike            	String  	执行价格
        callOI            	String  	看涨持仓总量（以币为单位）
        putOI             	String  	看跌持仓总量（以币为单位）
        callVol           	String  	看涨交易总量（以币为单位）
        putVol            	String  	看跌交易总量（以币为单位）
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_open_interest_volume_strike, **to_local(locals()))

    # 看跌/看涨期权合约 主动买入/卖出量
    def get_taker_block_volume(self, ccy: str, period: str = '', proxies={}, proxy_host: str = None):
        '''
        该指标展示某一时刻，单位时间内看跌/看涨期权的主动（taker）买入/卖出交易量
        https://www.okx.com/docs-v5/zh/#trading-statistics-rest-api-get-taker-flow
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	是       	币种
        period            	String  	否       	时间粒度，默认值8H。支持[8H/1D]每个粒度仅展示最新的一份数据
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	数据产生时间
        callBuyVol        	String  	看涨买入量  以结算货币为单位
        callSellVol       	String  	看涨卖出量  以结算货币为单位
        putBuyVol         	String  	看跌买入量  以结算货币为单位
        putSellVol        	String  	看跌卖出量  以结算货币为单位
        callBlockVol      	String  	看涨大单
        putBlockVol       	String  	看跌大单
        '''
        return self.send_request(*_TradingStatisticsEndpoints.get_taker_block_volume, **to_local(locals()))



'''
赚币
https://www.okx.com/docs-v5/zh/#financial-product-earn
'''
from paux.param import to_local
from okx.api._client import Client


class _EarnEndpoints():
    get_offers = ['/api/v5/finance/staking-defi/offers', 'GET']  # GET / 查看项目
    set_purchase = ['/api/v5/finance/staking-defi/purchase', 'POST']  # POST / 申购项目
    set_redeem = ['/api/v5/finance/staking-defi/redeem', 'POST']  # POST / 赎回项目
    set_cancel = ['/api/v5/finance/staking-defi/cancel', 'POST']  # POST / 撤销项目申购/赎回
    get_orders_active = ['/api/v5/finance/staking-defi/orders-active', 'GET']  # GET / 查看活跃订单
    get_orders_history = ['/api/v5/finance/staking-defi/orders-history', 'GET']  # GET / 查看历史订单


class Earn(Client):

    # GET / 查看项目
    def get_offers(self, productId: str = '', protocolType: str = '', ccy: str = '', proxies={},
                   proxy_host: str = None):
        '''
        GET /api/v5/finance/staking-defi/offers
        https://www.okx.com/docs-v5/zh/#financial-product-earn-get-offers
        
        限速：3次/s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        productId         	String  	否       	项目ID
        protocolType      	String  	否       	项目类型staking：锁仓挖矿defi：DEFI
        ccy               	String  	否       	投资币种，如BTC
        返回参数:
        Parameter         	Type    	Description
        ccy               	String  	币种名称，如BTC
        productId         	String  	项目ID
        protocol          	String  	项目名称
        protocolType      	String  	项目类型staking：锁仓挖矿defi：DEFI
        term              	String  	项目期限活期为0，其他则显示定期天数
        apy               	String  	预估年化如果年化为7% ，则该字段为0.07
        earlyRedeem       	Boolean 	项目是否支持提前赎回
        investData        	Array   	目前用户可用来投资的目标币种信息
        > ccy             	String  	投资币种，如BTC
        > bal             	String  	可投数量
        > minAmt          	String  	最小申购量
        > maxAmt          	String  	最大申购量
        earningData       	Array   	收益信息
        > ccy             	String  	收益币种，如BTC
        > earningType     	String  	收益类型0: 预估收益1: 累计发放收益
        state             	String  	项目状态purchasable：可申购sold_out：售罄stop：暂停申购
        '''
        return self.send_request(*_EarnEndpoints.get_offers, **to_local(locals()))

    # POST / 申购项目
    def set_purchase(self, productId: str, investData: list, term: str = '', tag: str = '', proxies={},
                     proxy_host: str = None):
        '''
        POST /api/v5/finance/staking-defi/purchase
        https://www.okx.com/docs-v5/zh/#financial-product-earn-post-purchase
        
        限速：2次/s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        productId         	String  	是       	项目ID
        investData        	Array   	是       	投资信息
        > ccy             	String  	是       	投资币种，如BTC
        > amt             	String  	是       	投资数量
        term              	String  	可选      	投资期限定期项目必须指定投资期限
        tag               	String  	否       	订单标签字母（区分大小写）与数字的组合，可以是纯字母、纯数字，且长度在1-16位之间
        返回参数:
        Parameter         	Type    	Description
        ordId             	String  	订单ID
        tag               	String  	订单标签
        '''
        return self.send_request(*_EarnEndpoints.set_purchase, **to_local(locals()))

    # POST / 赎回项目
    def set_redeem(self, ordId: str, protocolType: str, allowEarlyRedeem: bool = '', proxies={},
                   proxy_host: str = None):
        '''
        POST /api/v5/finance/staking-defi/redeem
        https://www.okx.com/docs-v5/zh/#financial-product-earn-post-redeem
        
        限速：2次/s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        ordId             	String  	是       	订单ID
        protocolType      	String  	是       	项目类型staking：锁仓挖矿defi：DEFI
        allowEarlyRedeem  	Boolean 	否       	是否提前赎回默认为false
        返回参数:
        Parameter         	Type    	Description
        ordId             	String  	订单ID
        tag               	String  	订单标签
        '''
        return self.send_request(*_EarnEndpoints.set_redeem, **to_local(locals()))

    # POST / 撤销项目申购/赎回
    def set_cancel(self, ordId: str, protocolType: str, proxies={}, proxy_host: str = None):
        '''
        POST /api/v5/finance/staking-defi/cancel
        https://www.okx.com/docs-v5/zh/#financial-product-earn-post-cancel-purchases-redemptions
        
        限速：2次/s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        ordId             	String  	是       	订单ID
        protocolType      	String  	是       	项目类型staking：锁仓挖矿defi：DEFI
        返回参数:
        Parameter         	Type    	Description
        ordId             	String  	订单ID
        tag               	String  	订单标签
        '''
        return self.send_request(*_EarnEndpoints.set_cancel, **to_local(locals()))

    # GET / 查看活跃订单
    def get_orders_active(self, productId: str = '', protocolType: str = '', ccy: str = '', state: str = '', proxies={},
                          proxy_host: str = None):
        '''
        GET /api/v5/finance/staking-defi/orders-active
        https://www.okx.com/docs-v5/zh/#financial-product-earn-get-active-orders
        
        限速：3次/s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        productId         	String  	否       	项目ID
        protocolType      	String  	否       	项目类型staking：锁仓挖矿defi：DEFI
        ccy               	String  	否       	投资币种，如BTC
        state             	String  	否       	订单状态8: 待上车（预约中）13: 订单取消中9: 上链中1: 收益中2: 赎回中
        返回参数:
        Parameter         	Type    	Description
        ccy               	String  	币种名称，如BTC
        ordId             	String  	订单ID
        productId         	String  	项目ID
        state             	String  	订单状态8: 待上车（预约中）13: 订单取消中9: 上链中1: 收益中2: 赎回中
        protocol          	String  	项目名称
        protocolType      	String  	项目类型staking：锁仓挖矿defi：DEFI
        term              	String  	项目期限活期为0，其他则显示定期天数
        apy               	String  	预估年化如果年化为7% ，则该字段为0.07保留到小数点后4位（截位）
        investData        	Array   	用户投资信息
        > ccy             	String  	投资币种，如BTC
        > amt             	String  	已投资数量
        earningData       	Array   	收益信息
        > ccy             	String  	收益币种，如BTC
        > earningType     	String  	收益类型0: 预估收益1: 实际到账收益
        > earnings        	String  	收益数量
        purchasedTime     	String  	用户订单创建时间，值为时间戳，Unix时间戳为毫秒数格式，如1597026383085
        purchasedTime     	String  	用户订单创建时间，值为时间戳，Unix时间戳为毫秒数格式，如1597026383085
        estSettlementTime 	String  	预估赎回到账时间
        cancelRedemptionDeadline	String  	撤销赎回申请截止时间
        tag               	String  	订单标签
        '''
        return self.send_request(*_EarnEndpoints.get_orders_active, **to_local(locals()))

    # GET / 查看历史订单
    def get_orders_history(self, productId: str = '', protocolType: str = '', ccy: str = '', after: str = '',
                           before: str = '', limit: str = '', proxies={}, proxy_host: str = None):
        '''
        GET /api/v5/finance/staking-defi/orders-history
        https://www.okx.com/docs-v5/zh/#financial-product-earn-get-order-history
        
        限速：3次/s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        productId         	String  	否       	项目ID
        protocolType      	String  	否       	项目类型staking：锁仓挖矿defi：DEFI
        ccy               	String  	否       	投资币种，如BTC
        after             	String  	否       	请求此ID之前（更旧的数据）的分页内容，传的值为对应接口的ordId
        before            	String  	否       	请求此ID之后（更新的数据）的分页内容，传的值为对应接口的ordId
        limit             	String  	否       	返回结果的数量，默认100条,最大值为100条
        返回参数:
        Parameter         	Type    	Description
        ccy               	String  	币种名称，如BTC
        ordId             	String  	订单ID
        productId         	String  	项目ID
        state             	String  	订单状态3: 订单已完成（包含撤销和已赎回两种状态）
        protocol          	String  	项目名称
        protocolType      	String  	项目类型staking：锁仓挖矿defi：DEFI
        term              	String  	项目期限活期为0，其他则显示定期天数
        apy               	String  	预估年化如果年化为7% ，则该字段为0.07保留到小数点后4位（截位）
        investData        	Array   	用户投资信息
        > ccy             	String  	投资币种，如BTC
        > amt             	String  	已投资数量
        earningData       	Array   	收益信息
        > ccy             	String  	收益币种，如BTC
        > earningType     	String  	收益类型0:预估收益1:实际到账收益
        > realizedEarnings	String  	已赎回订单累计收益该字段只在订单处于赎回状态时有效
        purchasedTime     	String  	用户订单创建时间，值为时间戳，Unix时间戳为毫秒数格式，如1597026383085
        redeemedTime      	String  	用户订单赎回时间，值为时间戳，Unix时间戳为毫秒数格式，如1597026383085
        tag               	String  	订单标签
        '''
        return self.send_request(*_EarnEndpoints.get_orders_history, **to_local(locals()))



'''
行情数据
https://www.okx.com/docs-v5/zh/#order-book-trading-market-data
'''
from paux.param import to_local
from okx.api._client import Client


class _MarketEndpoints():
    get_tickers = ['/api/v5/market/tickers', 'GET']  # GET / 获取所有产品行情信息
    get_ticker = ['/api/v5/market/ticker', 'GET']  # GET / 获取单个产品行情信息
    get_books = ['/api/v5/market/books', 'GET']  # GET / 获取产品深度
    get_books_lite = ['/api/v5/market/books-lite', 'GET']  # GET / 获取产品轻量深度
    get_candles = ['/api/v5/market/candles', 'GET']  # GET / 获取交易产品K线数据
    get_history_candles = ['/api/v5/market/history-candles', 'GET']  # GET / 获取交易产品历史K线数据
    get_trades = ['/api/v5/market/trades', 'GET']  # GET / 获取交易产品公共成交数据
    get_history_trades = ['/api/v5/market/history-trades', 'GET']  # GET / 获取交易产品公共历史成交数据
    get_instrument_family_trades = ['/api/v5/market/option/instrument-family-trades', 'GET']  # GET / 获取期权品种公共成交数据
    get_option_trades = ['/api/v5/public/option-trades', 'GET']  # GET / 获取期权公共成交数据
    get_platform_24_volume = ['/api/v5/market/platform-24-volume', 'GET']  # GET / 获取平台24小时总成交量


class Market(Client):

    # GET / 获取所有产品行情信息
    def get_tickers(self, instType: str, uly: str = '', instFamily: str = '', proxies={}, proxy_host: str = None):
        '''
        获取产品行情信息
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-tickers
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型SPOT：币币SWAP：永续合约FUTURES：交割合约OPTION：期权
        uly               	String  	否       	标的指数适用于交割/永续/期权，如BTC-USD
        instFamily        	String  	否       	交易品种适用于交割/永续/期权，如BTC-USD
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型
        instId            	String  	产品ID
        last              	String  	最新成交价
        lastSz            	String  	最新成交的数量
        askPx             	String  	卖一价
        askSz             	String  	卖一价的挂单数数量
        bidPx             	String  	买一价
        bidSz             	String  	买一价的挂单数量
        open24h           	String  	24小时开盘价
        high24h           	String  	24小时最高价
        low24h            	String  	24小时最低价
        volCcy24h         	String  	24小时成交量，以币为单位如果是衍生品合约，数值为交易货币的数量。如果是币币/币币杠杆，数值为计价货币的数量。
        vol24h            	String  	24小时成交量，以张为单位如果是衍生品合约，数值为合约的张数。如果是币币/币币杠杆，数值为交易货币的数量。
        sodUtc0           	String  	UTC 0  时开盘价
        sodUtc8           	String  	UTC+8 时开盘价
        ts                	String  	ticker数据产生时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_MarketEndpoints.get_tickers, **to_local(locals()))

    # GET / 获取单个产品行情信息
    def get_ticker(self, instId: str, proxies={}, proxy_host: str = None):
        '''
        获取产品行情信息
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-ticker
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USD-SWAP
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型
        instId            	String  	产品ID
        last              	String  	最新成交价
        lastSz            	String  	最新成交的数量
        askPx             	String  	卖一价
        askSz             	String  	卖一价对应的数量
        bidPx             	String  	买一价
        bidSz             	String  	买一价对应的数量
        open24h           	String  	24小时开盘价
        high24h           	String  	24小时最高价
        low24h            	String  	24小时最低价
        volCcy24h         	String  	24小时成交量，以币为单位如果是衍生品合约，数值为交易货币的数量。如果是币币/币币杠杆，数值为计价货币的数量。
        vol24h            	String  	24小时成交量，以张为单位如果是衍生品合约，数值为合约的张数。如果是币币/币币杠杆，数值为交易货币的数量。
        sodUtc0           	String  	UTC 0  时开盘价
        sodUtc8           	String  	UTC+8 时开盘价
        ts                	String  	ticker数据产生时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_MarketEndpoints.get_ticker, **to_local(locals()))

    # GET / 获取产品深度
    def get_books(self, instId: str, sz: str = '', proxies={}, proxy_host: str = None):
        '''
        获取产品深度列表
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-order-book
        
        限速：40次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USDT
        sz                	String  	否       	深度档位数量，最大值可传400，即买卖深度共800条不填写此参数，默认返回1档深度数据
        返回参数:
        Parameter         	Type    	Description
        asks              	Array   	卖方深度
        bids              	Array   	买方深度
        ts                	String  	深度产生的时间
        '''
        return self.send_request(*_MarketEndpoints.get_books, **to_local(locals()))

    # GET / 获取产品轻量深度
    def get_books_lite(self, instId: str, proxies={}, proxy_host: str = None):
        '''
        可以更快地获取前25档的深度信息。
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-order-lite-book
        
        限速：6次/1s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USDT
        返回参数:
        Parameter         	Type    	Description
        asks              	Array   	卖方深度
        bids              	Array   	买方深度
        ts                	String  	深度产生的时间
        '''
        return self.send_request(*_MarketEndpoints.get_books_lite, **to_local(locals()))

    # GET / 获取交易产品K线数据
    def get_candles(self, instId: str, bar: str = '', after: str = '', before: str = '', limit: str = '', proxies={},
                    proxy_host: str = None):
        '''
        获取K线数据。K线数据按请求的粒度分组返回，K线数据每个粒度最多可获取最近1,440条。
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-candlesticks
        
        限速：40次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USD-190927-5000-C
        bar               	String  	否       	时间粒度，默认值1m如 [1m/3m/5m/15m/30m/1H/2H/4H]香港时间开盘价k线：[6H/12H/1D/2D/3D/1W/1M/3M]UTC时间开盘价k线：[/6Hutc/12Hutc/1Dutc/2Dutc/3Dutc/1Wutc/1Mutc/3Mutc]
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts, 单独使用时，会返回最新的数据。
        limit             	String  	否       	分页返回的结果集数量，最大为300，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	开始时间，Unix时间戳的毫秒数格式，如1597026383085
        o                 	String  	开盘价格
        h                 	String  	最高价格
        l                 	String  	最低价格
        c                 	String  	收盘价格
        vol               	String  	交易量，以张为单位如果是衍生品合约，数值为合约的张数。如果是币币/币币杠杆，数值为交易货币的数量。
        volCcy            	String  	交易量，以币为单位如果是衍生品合约，数值为交易货币的数量。如果是币币/币币杠杆，数值为计价货币的数量。
        volCcyQuote       	String  	交易量，以计价货币为单位如：BTC-USDT 和 BTC-USDT-SWAP, 单位均是 USDT；BTC-USD-SWAP 单位是 USD
        confirm           	String  	K线状态0代表 K 线未完结，1代表 K 线已完结。
        '''
        return self.send_request(*_MarketEndpoints.get_candles, **to_local(locals()))

    # GET / 获取交易产品历史K线数据
    def get_history_candles(self, instId: str, after: str = '', before: str = '', bar: str = '', limit: str = '',
                            proxies={}, proxy_host: str = None):
        '''
        获取最近几年的历史k线数据(1s k线支持查询最近3个月的数据)
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-candlesticks-history
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USD-200927
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts, 单独使用时，会返回最新的数据。
        bar               	String  	否       	时间粒度，默认值1m如 [1s/1m/3m/5m/15m/30m/1H/2H/4H]香港时间开盘价k线：[6H/12H/1D/2D/3D/1W/1M/3M]UTC时间开盘价k线：[6Hutc/12Hutc/1Dutc/2Dutc/3Dutc/1Wutc/1Mutc/3Mutc]
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	开始时间，Unix时间戳的毫秒数格式，如1597026383085
        o                 	String  	开盘价格
        h                 	String  	最高价格
        l                 	String  	最低价格
        c                 	String  	收盘价格
        vol               	String  	交易量，以张为单位如果是衍生品合约，数值为合约的张数。如果是币币/币币杠杆，数值为交易货币的数量。
        volCcy            	String  	交易量，以币为单位如果是衍生品合约，数值为交易货币的数量。如果是币币/币币杠杆，数值为计价货币的数量。
        volCcyQuote       	String  	交易量，以计价货币为单位如：BTC-USDT 和 BTC-USDT-SWAP, 单位均是 USDT；BTC-USD-SWAP 单位是 USD
        confirm           	String  	K线状态0代表 K 线未完结，1代表 K 线已完结。
        '''
        return self.send_request(*_MarketEndpoints.get_history_candles, **to_local(locals()))

    # GET / 获取交易产品公共成交数据
    def get_trades(self, instId: str, limit: str = '', proxies={}, proxy_host: str = None):
        '''
        查询市场上的成交信息数据
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-trades
        
        限速：100次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USDT
        limit             	String  	否       	分页返回的结果集数量，最大为500，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        instId            	String  	产品ID
        tradeId           	String  	成交ID
        px                	String  	成交价格
        sz                	String  	成交数量
        side              	String  	成交方向buy：买sell：卖
        ts                	String  	成交时间，Unix时间戳的毫秒数格式， 如1597026383085
        '''
        return self.send_request(*_MarketEndpoints.get_trades, **to_local(locals()))

    # GET / 获取交易产品公共历史成交数据
    def get_history_trades(self, instId: str, type: str = '', after: str = '', before: str = '', limit: str = '',
                           proxies={}, proxy_host: str = None):
        '''
        查询市场上的成交信息数据，可以分页获取最近3个月的数据。
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-trades-history
        
        限速：10次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USDT
        type              	String  	否       	分页类型1：tradeId 分页2：时间戳分页默认为1：tradeId 分页
        after             	String  	否       	请求此 ID 或 ts 之前的分页内容，传的值为对应接口的 tradeId 或 ts
        before            	String  	否       	请求此ID之后（更新的数据）的分页内容，传的值为对应接口的 tradeId。不支持时间戳分页。单独使用时，会返回最新的数据。
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        instId            	String  	产品ID
        tradeId           	String  	成交ID
        px                	String  	成交价格
        sz                	String  	成交数量
        side              	String  	成交方向buy：买sell：卖
        ts                	String  	成交时间，Unix时间戳的毫秒数格式， 如1597026383085
        '''
        return self.send_request(*_MarketEndpoints.get_history_trades, **to_local(locals()))

    # GET / 获取期权品种公共成交数据
    def get_instrument_family_trades(self, instFamily: str, proxies={}, proxy_host: str = None):
        '''
        查询期权同一个交易品种下的成交信息数据，最多返回100条。
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-option-trades-by-instrument-family
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instFamily        	String  	是       	交易品种，如 BTC-USD，适用于期权
        返回参数:
        Parameter         	Type    	Description
        vol24h            	String  	24小时成交量，以张为单位
        optType           	String  	期权类型，C：看涨期权P：看跌期权
        tradeInfo         	Array   	成交数据列表
        > instId          	String  	产品ID
        > tradeId         	String  	成交ID
        > px              	String  	成交价格
        > sz              	String  	成交数量
        > side            	String  	成交方向buy：买sell：卖
        > ts              	String  	成交时间，Unix时间戳的毫秒数格式， 如1597026383085
        '''
        return self.send_request(*_MarketEndpoints.get_instrument_family_trades, **to_local(locals()))

    # GET / 获取期权公共成交数据
    def get_option_trades(self, instId: str = '', instFamily: str = '', optType: str = '', proxies={},
                          proxy_host: str = None):
        '''
        最多返回最近的100条成交数据
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-option-trades
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	可选      	产品ID，如 BTC-USD-221230-4000-C，instId和instFamily必须传一个，若传两个，以instId为主
        instFamily        	String  	可选      	交易品种，如 BTC-USD
        optType           	String  	否       	期权类型，C：看涨期权P：看跌期权
        返回参数:
        Parameter         	Type    	Description
        instId            	String  	产品ID
        instFamily        	String  	交易品种
        tradeId           	String  	成交ID
        px                	String  	成交价格
        sz                	String  	成交数量
        side              	String  	成交方向buy：买sell：卖
        optType           	String  	期权类型，C：看涨期权 P：看跌期权 ，仅适用于期权
        fillVol           	String  	成交时的隐含波动率（对应成交价格）
        fwdPx             	String  	成交时的远期价格
        idxPx             	String  	成交时的指数价格
        markPx            	String  	成交时的标记价格
        ts                	String  	成交时间，Unix时间戳的毫秒数格式， 如1597026383085
        '''
        return self.send_request(*_MarketEndpoints.get_option_trades, **to_local(locals()))

    # GET / 获取平台24小时总成交量
    def get_platform_24_volume(self, proxies={}, proxy_host: str = None):
        '''
        24小时成交量滚动计算
        https://www.okx.com/docs-v5/zh/#order-book-trading-market-data-get-24h-total-volume
        
        限速：2次/2s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        volUsd            	String  	订单簿交易近24小时总成交量，以美元为单位
        volCny            	String  	订单簿交易近24小时总成交量，以人民币为单位
        ts                	String  	接口返回数据时间
        '''
        return self.send_request(*_MarketEndpoints.get_platform_24_volume, **to_local(locals()))



import hmac
import base64
import datetime
import urllib.parse as up
import json
import requests
import time


# 请求方法错误
class RequestMethodError(Exception):
    error_msg: str

    def __init__(self, msg):
        self.error_msg = msg

    def __str__(self):
        return self.error_msg


# 响应状态码错误
class ResponseStatusError(Exception):
    error_msg: str

    def __init__(self, msg):
        self.error_msg = msg

    def __str__(self):
        return self.error_msg


def request_retry_wrapper(retry_num=50, retry_delay=0.1):
    def wrapper1(func):
        def wrapper2(*args, **kwargs):
            for i in range(retry_num - 1):
                try:
                    result = func(*args, **kwargs)
                    # 50001 服务暂时不可用，请稍后重试
                    # 50004 接口请求超时（不代表请求成功或者失败，请检查请求结果）
                    # 50011 用户请求频率过快，超过该接口允许的限额。请参考 API 文档并限制请求
                    # 50013 当前系统繁忙，请稍后重试
                    # 50026 系统错误，请稍后重试
                    if result['code'] in ['50001', '50004', '50011', '50013', '50026']:
                        time.sleep(retry_delay)
                        continue
                    else:
                        return result
                except requests.exceptions.ConnectionError:
                    time.sleep(retry_delay)
            result = func(*args, **kwargs)
            return result

        return wrapper2

    return wrapper1


class Client(object):
    API_URL = 'https://www.okex.com'

    def __init__(
            self,
            key='',
            secret='',
            passphrase='',
            flag='0',
            proxies={},
            proxy_host: str = None,
            retry_num=50,
            retry_delay=0.1,
    ):
        self.key = key
        self.secret = secret
        self.passphrase = passphrase
        self.flag = flag
        self.proxies = proxies
        if proxy_host and proxy_host.endswith('/'):
            proxy_host = proxy_host[0:-1]
        self.proxy_host = proxy_host
        self.retry_num = retry_num
        self.retry_delay = retry_delay

    def send_request(self, path, method, *args, **params):
        @request_retry_wrapper(retry_num=self.retry_num, retry_delay=self.retry_delay)
        def inner(path, method, **params):
            params_no_empty = {}
            for k, v in params.items():
                if v not in ['', [], {}, None] and k not in ['proxies', 'proxy_host']:
                    params_no_empty[k] = v
            # path
            if method == 'GET' and params_no_empty:
                path = path + '?' + up.urlencode(params_no_empty)
            # params中的proxy_host优先级 大于 self.proxy_host 大于 API_URL
            # url
            if params['proxy_host']:
                proxy_host = params['proxy_host']
                if proxy_host.endswith('/'):
                    proxy_host = proxy_host[0:-1]
                url = proxy_host + path
            elif self.proxy_host:
                url = self.proxy_host + path
            else:
                url = up.urljoin(self.API_URL, path)
            # params中proxies的优先级高于self.proxies
            if 'proxies' in params.keys() and params['proxies']:
                proxies = params['proxies']
            else:
                proxies = self.proxies
            # body
            if method == 'POST':
                # 类似批量下单这种无Key类型的body
                if args:
                    body = json.dumps(args[0])
                else:
                    body = json.dumps(params_no_empty)
            else:
                body = ''
            timestamp = datetime.datetime.utcnow().isoformat("T", "milliseconds") + 'Z'
            # sign
            sign = self._get_sign(
                message=self._pre_hash(timestamp, method, path, body),
                secret=self.secret,
            )
            # header
            header = self._get_header(
                key=self.key,
                sign=sign,
                passphrase=self.passphrase,
                flag=self.flag,
                timestamp=timestamp
            )

            # request
            if method == 'GET':
                if 'proxies' in params.keys() and params['proxies']:
                    response = requests.get(url, headers=header, proxies=proxies)
                else:
                    response = requests.get(url, headers=header, proxies=proxies)

            elif method == 'POST':
                response = requests.post(url, data=body, headers=header, proxies=self.proxies)
            else:
                msg = 'Error request method {method}'.format(method=str(method))
                raise RequestMethodError(msg)

            if not str(response.status_code).startswith('2'):
                msg = 'Error response_status_code {code}\nresponse_content={content}'.format(code=response.status_code,
                                                                                             content=response.text)
                raise ResponseStatusError(msg)
            return response.json()

        return inner(path, method, **params)

    @staticmethod
    def _get_sign(message, secret):
        mac = hmac.new(bytes(secret, encoding='utf8'), bytes(message, encoding='utf-8'), digestmod='sha256')
        d = mac.digest()
        return base64.b64encode(d)

    @staticmethod
    def _pre_hash(timestamp, method, request_path, body):
        return str(timestamp) + str.upper(method) + request_path + body

    @staticmethod
    def _get_header(key, sign, passphrase, flag, timestamp):
        header = {}
        header['Content-Type'] = 'application/json'
        header['OK-ACCESS-KEY'] = key
        header['OK-ACCESS-SIGN'] = sign
        header['OK-ACCESS-TIMESTAMP'] = str(timestamp)
        header['OK-ACCESS-PASSPHRASE'] = passphrase
        header['x-simulated-trading'] = flag
        return header



# 交易对信息错误
EXCHANGE_INFO_ERROR = ['EXCHANGE_INFO_ERROR', 'Transaction rules and transaction pair errors']
# 历史K线时间起点错误
CANDLE_START_ERROR = ['CANDLE_START_ERROR', 'History candle start ts error']
# 历史K线时间终点错误
CANDLE_END_ERROR = ['CANDLE_END_ERROR', 'History candle end ts error']
# 历史K线长度错误
CANDLE_LENGTH_ERROR = ['CANDLE_LENGTH_ERROR', 'History candle length error']
# 历史K线时间间隔错误
CANDLE_INTERVAL_ERROR = ['CANDLE_INTERVAL_ERROR', 'History candle ts interval error']



import time
import datetime
import pendulum
import pandas as pd
import numpy as np
import paux.date as _date
from typing import Union
from okx.app import code
from okx.app.market._base import MarketBase
from okx.app.market import utils
from candlelite.calculate import valid as _valid
from candlelite.calculate import transform as _transform
from candlelite.calculate import interval as _interval

__all__ = ['HistoryCandle']


# 历史K线
class HistoryCandle(MarketBase):
    # 将API返回的原始历史K线数据列表转换为np.array类型
    def __candle_list_to_candle(
            self,
            instId: str,
            candle_list: list,
            start: Union[int, float, str, datetime.date, datetime.datetime, None],
            end: Union[int, float, str, datetime.date, datetime.datetime, None],
            bar: str = '1m',
            valid_interval: bool = True,
            valid_start: bool = True,
            valid_end: bool = True
    ) -> dict:
        '''
        :param instId: 产品
        :param candle_list: API返回的K线数据列表
        :param start: 起始时间
        :param end: 终止时间
            start和end可以为None，如果存在数据，会截取[start,end]的数据
            例如：
                start == None end != None
                    -> ts <= end
                start != None end == None
                    -> ts >= start
                start != None end != None
                    -> start <= ts <= end
        :param bar: 时间粒度
        :param valid_interval: 是否验证时间间隔
        :param valid_start: 是否验证时间起点 start必须存在
        :param valid_end: 是否验证时间终点 end必须存在
        '''
        # 转换DataFrame
        df = pd.DataFrame(candle_list).astype(float)
        # [ERROR RETURN] 数据为空
        if not df.shape[0]:
            result = {
                'code': code.CANDLE_LENGTH_ERROR[0],
                'data': np.array([]),
                'msg': 'Candle empty {instId} {start} ~ {end}'.format(
                    instId=str(instId),
                    start=_date.to_fmt(date=start, timezone=self.timezone) if start else 'None',
                    end=_date.to_fmt(date=end, timezone=self.timezone) if end else 'None',
                ),
            }
            return result
        # 按照时间戳去重
        ts_column = df.columns.tolist()[0]
        df[ts_column] = df[ts_column].astype(int)
        df = df.drop_duplicates(subset=ts_column)
        # 按照start和end截取数据
        if start and end:
            df = df[
                (df[ts_column] <= _date.to_ts(end, timezone=self.timezone)) &
                (df[ts_column] >= _date.to_ts(start, timezone=self.timezone))
                ]
        elif start and not end:
            df = df[df[ts_column] >= _date.to_ts(start, timezone=self.timezone)]
        elif end and not start:
            df = df[df[ts_column] <= _date.to_ts(end, timezone=self.timezone)]
        # 排序
        df = df.sort_values(by=ts_column)
        # 转化为array
        candle = df.to_numpy()
        # 验证时间起点
        if valid_start:
            valid_start_result = _valid.valid_start(candle=candle, start=start, timezone=self.timezone)
            # [ERROR RETURN] 时间起点错误
            if not valid_start_result['code']:
                return {
                    'code': code.CANDLE_START_ERROR[0],
                    'data': candle,
                    'msg': str(instId) + valid_start_result['msg']
                }
        # 验证时间终点
        if valid_end:
            valid_end_result = _valid.valid_end(candle=candle, end=end, timezone=self.timezone)
            # [ERROR RETURN] 时间终点错误
            if not valid_end_result['code']:
                return {
                    'code': code.CANDLE_END_ERROR[0],
                    'data': candle,
                    'msg': str(instId) + valid_end_result['msg'],
                }

        # 验证时间间隔
        if valid_interval:
            valid_interval_result = _valid.valid_interval(candle=candle, interval=_interval.get_interval(bar))
            # [ERROR RETURN] 时间间隔错误
            if not valid_interval_result['code']:
                return {
                    'code': code.CANDLE_INTERVAL_ERROR[0],
                    'data': candle,
                    'msg': str(instId) + valid_interval_result['msg'],
                }
        # [RETURN]
        result = {'code': '0', 'data': candle, 'msg': ''}
        return result

    # 获取产品的历史K线数据
    # Weight >= 1 (智能节约权重)
    def get_history_candle(
            self,
            instId: str,
            start: Union[str, int, float, datetime.datetime, datetime.date],
            end: Union[str, int, float, datetime.datetime, datetime.date],
            bar: str = '1m',
            valid_interval: bool = True,
            valid_start: bool = True,
            valid_end: bool = True
    ) -> dict:
        '''
        :param instId: 产品
        :param start: 起始日期时间
        :param end: 起始日期时间
        :param bar: 时间粒度
        :param valid_interval: 是否验证时间间隔
        :param valid_start: 是否验证起始时间
        :param valid_end: 是否验证终止时间

        okx 底层API不会包含before（start）和after（end）时刻的数据
        '''
        # 起始与终止时间戳
        start_ts = _date.to_ts(date=start, timezone=self.timezone)
        end_ts = _date.to_ts(date=end, timezone=self.timezone)
        # 时间间隔
        bar_interval = _interval.get_interval(bar=bar)
        # 初始化起始时间和candle列表
        before_ts = start_ts
        candle_list = []
        while True:
            # 计算limit
            limit = self._get_limit(num=(end_ts - before_ts) / bar_interval + 1)
            # 获取K线数据
            result = self.marketAPI.get_history_candles(
                instId=instId,
                before=before_ts - 1,
                after=before_ts + (limit - 1) * bar_interval + 1,
                bar=bar,
                limit=limit,
            )
            # [ERROR RETURN] K线数据错误
            if result['code'] != '0':
                return result
            # [BREAK] 数据为空
            if not result['data']:
                break
            # 保存candle
            candle_list += result['data']
            # 跳过上一次的endTime，上一次的endTime=before_ts + (limit - 1) * bar_interval
            before_ts = before_ts + limit * bar_interval
            # [BREAK] 完成
            if before_ts > end_ts:
                break
        # 转换为candle
        result = self.__candle_list_to_candle(
            instId=instId,
            candle_list=candle_list,
            start=start,
            end=end,
            bar=bar,
            valid_interval=valid_interval,
            valid_start=valid_start,
            valid_end=valid_end,
        )
        # [RETURN]
        return result

    # 获取产品指定数量的最新历史K线数据
    # Weight > 1 (智能节约权重)
    def get_history_candle_latest(
            self,
            instId: str,
            length: int,
            end: Union[str, int, float, datetime.date, datetime.datetime, None] = None,
            bar: str = '1m',
            valid_interval: bool = True,
    ) -> dict:
        '''
        :param instId: 合约产品
        :param length: K线长度
        :param end: 终止日期时间
            支持: 毫秒时间戳、日期格式的字符串、日期时间对象与日期对象 -> 自动验证数据终点
            None: 取当前时间 -> 不验证时间终点
        :param bar: 时间粒度
        :param valid_interval: 是否验证时间间隔
        '''
        # 时间终止
        if not end:
            end_ts = int(time.time() * 1000)
        else:
            end_ts = _date.to_ts(date=end, timezone=self.timezone)
        # 时间间隔
        bar_interval = _interval.get_interval(bar)
        # 从后向前获取candle
        after_ts = end_ts
        candle_list = []
        while True:
            # 计算limit
            limit = self._get_limit(num=length - len(candle_list))
            # 获取K线数据
            result = self.marketAPI.get_history_candles(
                instId=instId,
                before=after_ts - (limit - 1) * bar_interval - 1,
                after=after_ts + 1,
                bar=bar,
                limit=limit,
            )
            # [ERROR RETURN] 数据错误
            if result['code'] != '0':
                return result
            # [BREAK] 数据为空
            if not result['data']:
                break
            # 添加
            candle_list += result['data']
            # [BREAK] 完成
            if len(candle_list) >= length:
                break
            # 计算下一个的终止时间
            after_ts = int(candle_list[-1][0]) - bar_interval
        # 转换为np.array
        result = self.__candle_list_to_candle(
            instId=instId,
            candle_list=candle_list,
            start=None,
            end=end,
            bar=bar,
            valid_interval=valid_interval,
            valid_start=False,
            valid_end=True if end != None else False,  # 是否验证终止时间
        )
        # [ERROR RETURN] 转换失败
        if result['code'] != '0':
            return result
        # 取后length行数据
        candle = result['data'][-length:]
        # 验证数据长度
        valid_length_result = _valid.valid_length(candle=candle, length=length)
        # [ERROR RETURN] 数据长度错误
        if not valid_length_result['code']:
            return {
                'code': code.CANDLE_LENGTH_ERROR[0],
                'data': candle,
                'msg': str(instId) + valid_length_result['msg']
            }
        # [RETURN]
        result['data'] = candle
        return result

    # 获取产品指定日期的历史K线数据
    # Weight: 1 ~ 10 (智能节约权重)
    def get_history_candle_by_date(
            self,
            instId: str,
            date: Union[str, int, float, datetime.date],
            bar: str = '1m',
            valid_interval: bool = True,
            valid_start: bool = True,
            valid_end: bool = True
    ) -> dict:
        '''
        :param instId: 产品
        :param date: 日期
        :param bar: 时间粒度
        :param valid_interval: 是否验证时间间隔
        :param valid_start: 是否验证起始时间
        :param valid_end: 是否验证终止时间
        '''
        # 日期时间对象，其中小时分钟秒均为0
        date = _date.to_datetime(date=date, timezone=self.timezone)
        # 起始毫秒时间戳
        start = int(date.timestamp() * 1000)
        bar_interval = _interval.get_interval(bar)
        # 终止毫秒时间戳
        end_date = datetime.date(year=date.year, month=date.month, day=date.day) + datetime.timedelta(days=1)
        end_date = pendulum.datetime(
            year=end_date.year,
            month=end_date.month,
            day=end_date.day,
            hour=0,
            minute=0,
            second=0,
            tz=self.timezone,
        )
        end = int(end_date.timestamp() * 1000 - 1 * bar_interval)
        # [RETURN]
        return self.get_history_candle(
            instId=instId,
            start=start,
            bar=bar,
            end=end,
            valid_interval=valid_interval,
            valid_start=valid_start,
            valid_end=valid_end,
        )

    # 获取历史K线数据中最新的毫秒时间戳
    # Weight: 1
    def get_history_candle_latest_ts(
            self,
            bar: str = '1m',
    ) -> dict:
        '''
        :param bar: 时间粒度
        '''
        instId = 'BTC-USDT'
        # 获取长度为1的最新历史K数据
        result = self.get_history_candle_latest(
            instId=instId,
            bar=bar,
            length=99,
        )
        # [ERROR RETURN] 历史K线有误
        if result['code'] != '0':
            return result
        # 最新的毫秒时间戳，以浮点数保存
        latest_ts = float(result['data'][-1, 0])
        result['data'] = latest_ts
        # [RETURN]
        return result

    # 更新产品历史K线数据到指定时间
    # Weight: >= 1 (Intelligent saving weight)
    def update_history_candle(
            self,
            instId: str,
            length: int,
            candle: np.array = None,
            end: [int, float, datetime.date, datetime.datetime, str] = None,
            bar: str = '1m',
            valid_interval=True,
    ) -> dict:
        '''
        :param instId: 合约产品
        :param length: 更新后的有效长度
        :param candle: 历史K线数据 None: 重新下载candle数据
        :param end: 更新后的终点时间
            None: 使用最新的历史K线数据毫秒时间戳
        :param bar: 时间粒度
        :param valid_interval: 是否验证时间间隔

        如论end是有有值，都会验证数据的时间终点与长度
        '''
        # 终止时间戳end_ts
        if end:
            # end有值
            end_ts = _date.to_ts(date=end, timezone=self.timezone)
        else:
            # 获取最新的历史K线时间戳
            result = self.get_history_candle_latest_ts(
                bar=bar,
            )
            # [ERROR RETURN] 获取时间戳异常
            if result['code'] != '0':
                return result
            end_ts = result['data']
        # 时间间隔
        bar_interval = _interval.get_interval(bar)
        # 计算剩余需要更新的长度 length_left
        # 存在candle
        if isinstance(candle, np.ndarray) and candle.shape[0] > 0:
            # 数据间隔异常 更新全部
            if not (np.diff(candle[:, 0]) == bar_interval).all():
                length_left = length
            else:
                # 数据过早 更新全部
                if end_ts - (length - 1) * bar_interval < candle[:, 0].min():
                    length_left = length
                # 更新剩余部分
                else:
                    candle_end_ts = candle[:, 0].max()
                    length_left = int(min((end_ts - candle_end_ts) / bar_interval, length))
        # 不存在candle，更新全部
        else:
            length_left = length
        # [RETURN] 无需更新
        if length_left == 0:
            return {'code': '0', 'data': candle, 'msg': ''}
        # 获取最新数据
        result = self.get_history_candle_latest(
            instId=instId,
            length=length_left,
            end=end_ts,
            bar=bar,
            valid_interval=valid_interval,
        )
        # [ERROR RETURN] 最新历史K线数据异常
        if result['code'] != '0':
            return result
        # 如果candle存在，二者拼接 candle_union
        candle_add = result['data']
        if isinstance(candle, np.ndarray) and candle.shape[0] > 0:
            candle_union = _transform.concat_candle(
                candles=[candle, candle_add],
                drop_duplicate=True,
                sort=True
            )
        else:
            candle_union = candle_add
        # 取末尾length长度
        candle_union = candle_union[-length:]
        # 验证时间间隔
        if valid_interval:
            valid_interval_result = _valid.valid_interval(candle=candle_union, interval=_interval.get_interval(bar))
            # [ERROR RETURN] 时间间隔错误
            if not valid_interval_result['code']:
                return {
                    'code': code.CANDLE_INTERVAL_ERROR[0],
                    'data': candle_union,
                    'msg': str(instId) + valid_interval_result['msg']
                }
        # 验证数据长度
        valid_length_result = _valid.valid_length(candle=candle_union, length=length)
        # 【ERROR RETURN】数据长度错误
        if not valid_length_result['code']:
            return {
                'code': code.CANDLE_LENGTH_ERROR[0],
                'data': candle_union,
                'msg': str(instId) + valid_length_result['msg'],
            }

        # [RETURN]
        result = {'code': '0', 'data': candle_union, 'msg': ''}
        return result

    # 历史K线 根据num获取最佳的limit
    def _get_limit(self, num: int) -> int:
        return int(np.clip(num, 1, 100))

    # 将Okx的历史K线数据从array转换成DataFrame类型
    def candle_to_df(self, candle: np.ndarray, convert_ts: bool = True) -> pd.DataFrame:
        return utils.candle_to_df(candle=candle, convert_ts=convert_ts, timezone=self.timezone)

    # 将Okx的历史K线数据从DataFrame转换成array类型
    def df_to_candle(self, df: pd.DataFrame, convert_ts: bool = True) -> np.ndarray:
        return utils.df_to_candle(df=df, convert_ts=convert_ts, timezone=self.timezone)



from okx.app.account import AccountSPOT
from okx.app.account import AccountSWAP
from okx.app.market import MarketSWAP
from okx.app.market import MarketSPOT
from okx.app.market import Market
from okx.app.trade.tradeSPOT import TradeSPOT
from okx.app.trade.tradeSWAP import TradeSWAP


class OkxSPOT():
    def __init__(self,
                 key: str,
                 secret: str,
                 passphrase: str,
                 timezone: str = 'Asia/Shanghai',
                 proxies={},
                 proxy_host: str = None,
                 ):
        self.account = AccountSPOT(
            key=key, secret=secret, passphrase=passphrase, proxies=proxies, proxy_host=proxy_host,
        )
        self.market = MarketSPOT(
            key=key, secret=secret, passphrase=passphrase, timezone=timezone, proxies=proxies, proxy_host=proxy_host,
        )
        self.trade = TradeSPOT(
            key=key, secret=secret, passphrase=passphrase,
            timezone=timezone,
            account=self.account,
            market=self.market,
            proxies=proxies, proxy_host=proxy_host,
        )
        self.timezone = timezone


class OkxSWAP():
    def __init__(self,
                 key: str,
                 secret: str,
                 passphrase: str,
                 timezone: str = 'Asia/Shanghai',
                 proxies={},
                 proxy_host: str = None,
                 ):
        self.account = AccountSWAP(
            key=key, secret=secret, passphrase=passphrase, proxies=proxies, proxy_host=proxy_host,
        )
        self.market = MarketSWAP(
            key=key, secret=secret, passphrase=passphrase, timezone=timezone, proxies=proxies, proxy_host=proxy_host,
        )
        self.trade = TradeSWAP(
            key=key, secret=secret, passphrase=passphrase,
            timezone=timezone,
            account=self.account,
            market=self.market,
            proxies=proxies,
            proxy_host=proxy_host,
        )
        self.timezone = timezone



from okx.app.exception._base import AbstractEXP


# 参数异常
class ParamException(AbstractEXP):
    pass


class ParamRoundPriceTypeException(AbstractEXP):
    def __init__(self, type, pmt='type must in ["CEIL","FLOOR"]'):
        self.error_msg = 'type = {type} \n{pmt}'.format(
            type=str(type),
            pmt=pmt,
        )


class ParamPositionSideException(AbstractEXP):
    def __init__(self, positionSide, pmt='positionSide must in ["LONG","SHORT"]'):
        self.error_msg = 'positionSide = {positionSide} \n{pmt}'.format(
            positionSide=str(positionSide),
            pmt=pmt,
        )



from okx.app.account import AccountSWAP
from okx.app.market import MarketSWAP
from okx.api import Trade as TradeAPI


class TradeBase():
    def __init__(
            self,
            key: str,
            secret: str,
            passphrase: str,
            timezone: str = 'Asia/Shanghai',
            account=None,
            market=None,
            proxies={},
            proxy_host: str = None,
    ):
        # SWAP账户
        if not account:
            self._account = AccountSWAP(key=key, secret=secret, passphrase=passphrase, proxies=proxies,
                                        proxy_host=proxy_host)
        else:
            self._account = account

        # SWAP行情
        if not market:
            self._market = MarketSWAP(key=key, secret=secret, passphrase=passphrase, timezone=timezone, proxies=proxies,
                                      proxy_host=proxy_host)
        else:
            self._market = market

        # 设置持仓方向
        set_position_mode_result = self._account.set_position_mode(
            posMode='long_short_mode'
        )
        if set_position_mode_result['code'] == '0':
            print('[SUCCESS] 设置持仓方式为双向持仓成功，posMode="long_short_mode"')
        else:
            print('[FAILURE] 设置持仓方式为双向持仓失败，请手动设置：posMode="long_short_mode"')

        # 时区
        self.timezone = timezone
        # TRADE API
        FLAG = '0'
        self.api = TradeAPI(key=key, secret=secret, passphrase=passphrase, flag=FLAG, proxies=proxies,
                            proxy_host=proxy_host)



'''
定投
https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy
'''
from paux.param import to_local
from okx.api._client import Client


class _RecurringBuyEndpoints():
    set_order_algo = ['/api/v5/tradingBot/recurring/order-algo', 'POST']  # POST / 定投策略委托下单
    set_amend_order_algo = ['/api/v5/tradingBot/recurring/amend-order-algo', 'POST']  # POST / 修改定投策略订单
    set_stop_order_algo = ['/api/v5/tradingBot/recurring/stop-order-algo', 'POST']  # POST / 定投策略停止
    get_orders_algo_pending = ['/api/v5/tradingBot/recurring/orders-algo-pending', 'GET']  # GET / 获取未完成定投策略委托单列表
    get_orders_algo_history = ['/api/v5/tradingBot/recurring/orders-algo-history', 'GET']  # GET / 获取历史定投策略委托单列表
    get_orders_algo_details = ['/api/v5/tradingBot/recurring/orders-algo-details', 'GET']  # GET / 获取定投策略委托订单详情
    get_sub_orders = ['/api/v5/tradingBot/recurring/sub-orders', 'GET']  # GET / 获取定投策略子订单信息


class RecurringBuy(Client):

    # POST / 定投策略委托下单
    def set_order_algo(self, stgyName: str, recurringList: object, period: str, recurringDay: str, recurringTime: str,
                       timeZone: str, amt: str, investmentCcy: str, tdMode: str, algoClOrdId: str = '', tag: str = '',
                       proxies={}, proxy_host: str = None):
        '''
        POST /api/v5/tradingBot/recurring/order-algo
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-post-place-recurring-buy-order
        
        限速：20次/2s
        
    
        请求参数:
        Parameter         	Type    	Required	Description

        stgyName          	String  	是       	策略自定义名称，不超过40个字符
        recurringList     	Array of object	是       	定投信息
        > ccy             	String  	是       	定投币种，如BTC
        > ratio           	String  	是       	定投币种资产占比，如 "0.2"代表占比20%
        period            	String  	是       	周期类型每月：monthly每周：weekly每日：daily
        recurringDay      	String  	是       	投资日当周期类型为monthly，则取值范围是 [1,28] 的整数当周期类型为weekly，则取值范围是 [1,7] 的整数当周期类型为daily，则取值为1
        recurringTime     	String  	是       	投资时间，取值范围是 [0,23] 的整数
        timeZone          	String  	是       	时区（UTC），取值范围是 [-12,14] 的整数如8表示UTC+8（东8区），北京时间
        amt               	String  	是       	每期投入数量
        investmentCcy     	String  	是       	投入数量单位，只能是USDT/USDC
        tdMode            	String  	是       	交易模式跨币种/PM保证金模式下选择 全仓：cross简单/单币种模式下选择 非保证金：cash
        algoClOrdId       	String  	否       	客户自定义订单ID字母（区分大小写）与数字的组合，可以是纯字母、纯数字且长度要在1-32位之间。
        tag               	String  	否       	订单标签字母（区分大小写）与数字的组合，可以是纯字母、纯数字，且长度在1-16位之间。
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        algoClOrdId       	String  	客户自定义订单ID
        sCode             	String  	事件执行结果的code，0代表成功
        sMsg              	String  	事件执行失败时的msg
        '''
        return self.send_request(*_RecurringBuyEndpoints.set_order_algo, **to_local(locals()))

    # POST / 修改定投策略订单
    def set_amend_order_algo(self, algoId: str, stgyName: str, proxies={}, proxy_host: str = None):
        '''
        POST /api/v5/tradingBot/recurring/amend-order-algo
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-post-amend-recurring-buy-order
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        algoId            	String  	是       	策略订单ID
        stgyName          	String  	是       	调整后的策略自定义名称
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        algoClOrdId       	String  	客户自定义订单ID
        sCode             	String  	事件执行结果的code，0代表成功
        sMsg              	String  	事件执行失败时的msg
        '''
        return self.send_request(*_RecurringBuyEndpoints.set_amend_order_algo, **to_local(locals()))

    # POST / 定投策略停止
    def set_stop_order_algo(self, algoId: str, proxies={}, proxy_host: str = None):
        '''
        每次最多可以撤销10个定投策略订单。
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-post-stop-recurring-buy-order
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        algoId            	String  	是       	策略订单ID
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        algoClOrdId       	String  	客户自定义订单ID
        sCode             	String  	事件执行结果的code，0代表成功
        sMsg              	String  	事件执行失败时的msg
        '''
        return self.send_request(*_RecurringBuyEndpoints.set_stop_order_algo, **to_local(locals()))

    # GET / 获取未完成定投策略委托单列表
    def get_orders_algo_pending(self, algoId: str = '', after: str = '', before: str = '', limit: str = '', proxies={},
                                proxy_host: str = None):
        '''
        GET /api/v5/tradingBot/recurring/orders-algo-pending
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-get-recurring-buy-order-list
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        algoId            	String  	否       	策略订单ID
        after             	String  	否       	请求此ID之前（更旧的数据）的分页内容，传的值为对应接口的algoId
        before            	String  	否       	请求此ID之后（更新的数据）的分页内容，传的值为对应接口的algoId
        limit             	String  	否       	返回结果的数量，最大为100，默认100条
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        algoClOrdId       	String  	客户自定义订单ID
        instType          	String  	产品类型SPOT：现货
        cTime             	String  	策略订单创建时间，Unix时间戳的毫秒数格式，如1597026383085
        uTime             	String  	策略订单更新时间，Unix时间戳的毫秒数格式，如1597026383085
        algoOrdType       	String  	策略订单类型recurring：定投
        state             	String  	订单状态running：运行中stopping：终止中
        stgyName          	String  	策略自定义名称，不超过40个字符
        recurringList     	Array of object	定投信息
        > ccy             	String  	定投币种，如BTC
        > ratio           	String  	定投币种资产占比，如 "0.2"代表占比20%
        period            	String  	周期类型每月：monthly每周：weekly每日：daily
        recurringDay      	String  	投资日当周期类型为monthly，则取值范围是 [1,28] 的整数当周期类型为weekly，则取值范围是 [1,7] 的整数当周期类型为daily，则取值为1
        recurringTime     	String  	投资时间，取值范围是 [0,23] 的整数
        timeZone          	String  	时区（UTC），取值范围是 [-12,14] 的整数如8表示UTC+8（东8区），北京时间
        amt               	String  	每期投入数量
        investmentAmt     	String  	累计投入数量
        investmentCcy     	String  	投入数量单位，只能是USDT/USDC
        totalPnl          	String  	总收益
        totalAnnRate      	String  	总年化
        pnlRatio          	String  	收益率
        mktCap            	String  	当前总市值，单位为USDT
        cycles            	String  	定投累计轮数
        tag               	String  	订单标签
        '''
        return self.send_request(*_RecurringBuyEndpoints.get_orders_algo_pending, **to_local(locals()))

    # GET / 获取历史定投策略委托单列表
    def get_orders_algo_history(self, algoId: str = '', after: str = '', before: str = '', limit: str = '', proxies={},
                                proxy_host: str = None):
        '''
        GET /api/v5/tradingBot/recurring/orders-algo-history
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-get-recurring-buy-order-history
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        algoId            	String  	否       	策略订单ID
        after             	String  	否       	请求此ID之前（更旧的数据）的分页内容，传的值为对应接口的algoId
        before            	String  	否       	请求此ID之后（更新的数据）的分页内容，传的值为对应接口的algoId
        limit             	String  	否       	返回结果的数量，最大为100，默认100条
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        algoClOrdId       	String  	客户自定义订单ID
        instType          	String  	产品类型SPOT：现货
        cTime             	String  	策略订单创建时间，Unix时间戳的毫秒数格式，如1597026383085
        uTime             	String  	策略订单更新时间，Unix时间戳的毫秒数格式，如1597026383085
        algoOrdType       	String  	策略订单类型recurring：定投
        state             	String  	订单状态stopped：已停止
        stgyName          	String  	策略自定义名称，不超过40个字符
        recurringList     	Array of object	定投信息
        > ccy             	String  	定投币种，如BTC
        > ratio           	String  	定投币种资产占比，如 "0.2"代表占比20%
        period            	String  	周期类型每月：monthly每周：weekly每日：daily
        recurringDay      	String  	投资日当周期类型为monthly，则取值范围是 [1,28] 的整数当周期类型为weekly，则取值范围是 [1,7] 的整数当周期类型为daily，则取值为1
        recurringTime     	String  	投资时间，取值范围是 [0,23] 的整数
        timeZone          	String  	时区（UTC），取值范围是 [-12,14] 的整数如8表示UTC+8（东8区），北京时间
        amt               	String  	每期投入数量
        investmentAmt     	String  	累计投入数量
        investmentCcy     	String  	投入数量单位，只能是USDT/USDC
        totalPnl          	String  	总收益
        totalAnnRate      	String  	总年化
        pnlRatio          	String  	收益率
        mktCap            	String  	当前总市值，单位为USDT
        cycles            	String  	定投累计轮数
        tag               	String  	订单标签
        '''
        return self.send_request(*_RecurringBuyEndpoints.get_orders_algo_history, **to_local(locals()))

    # GET / 获取定投策略委托订单详情
    def get_orders_algo_details(self, algoId: str, proxies={}, proxy_host: str = None):
        '''
        GET /api/v5/tradingBot/recurring/orders-algo-details
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-get-recurring-buy-order-details
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        algoId            	String  	是       	策略订单ID
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        algoClOrdId       	String  	客户自定义订单ID
        instType          	String  	产品类型SPOT：现货
        cTime             	String  	策略订单创建时间，Unix时间戳的毫秒数格式，如1597026383085
        uTime             	String  	策略订单更新时间，Unix时间戳的毫秒数格式，如1597026383085
        algoOrdType       	String  	策略订单类型recurring：定投
        state             	String  	订单状态running：运行中stopping：终止中stopped：已停止
        stgyName          	String  	策略自定义名称，不超过40个字符
        recurringList     	Array of object	定投信息
        > ccy             	String  	定投币种，如BTC
        > ratio           	String  	定投币种资产占比，如 "0.2"代表占比20%
        > totalAmt        	String  	累计购入定投币种的数量
        > profit          	String  	定投收益，单位为investmentCcy
        > avgPx           	String  	定投均价，计价单位为investmentCcy
        > px              	String  	当前价格，计价单位为investmentCcy
        period            	String  	周期类型每月：monthly每周：weekly每日：daily
        recurringDay      	String  	投资日当周期类型为monthly，则取值范围是 [1,28] 的整数当周期类型为weekly，则取值范围是 [1,7] 的整数当周期类型为daily，则取值为1
        recurringTime     	String  	投资时间，取值范围是 [0,23] 的整数
        timeZone          	String  	时区（UTC），取值范围是 [-12,14] 的整数如8表示UTC+8（东8区），北京时间
        amt               	String  	每期投入数量
        investmentAmt     	String  	累计投入数量
        investmentCcy     	String  	投入数量单位，只能是USDT/USDC
        totalPnl          	String  	总收益
        totalAnnRate      	String  	总年化
        pnlRatio          	String  	收益率
        mktCap            	String  	当前总市值，单位为USDT
        cycles            	String  	定投累计轮数
        tag               	String  	订单标签
        '''
        return self.send_request(*_RecurringBuyEndpoints.get_orders_algo_details, **to_local(locals()))

    # GET / 获取定投策略子订单信息
    def get_sub_orders(self, algoId: str, ordId: str = '', after: str = '', before: str = '', limit: str = '',
                       proxies={}, proxy_host: str = None):
        '''
        GET /api/v5/tradingBot/recurring/sub-orders
        https://www.okx.com/docs-v5/zh/#order-book-trading-recurring-buy-get-recurring-buy-sub-orders
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        algoId            	String  	是       	策略订单ID
        ordId             	String  	否       	子订单ID
        after             	String  	否       	请求此ID之前（更旧的数据）的分页内容，传的值为对应接口的ordId
        before            	String  	否       	请求此ID之后（更新的数据）的分页内容，传的值为对应接口的ordId
        limit             	String  	否       	返回结果的数量，最大为300，默认300条
        返回参数:
        Parameter         	Type    	Description
        algoId            	String  	策略订单ID
        instType          	String  	产品类型
        instId            	String  	产品ID
        algoOrdType       	String  	策略订单类型recurring：定投
        ordId             	String  	子订单ID
        cTime             	String  	子订单创建时间，Unix时间戳的毫秒数格式，如1597026383085
        uTime             	String  	子订单更新时间，Unix时间戳的毫秒数格式，如1597026383085
        tdMode            	String  	子订单交易模式cross：全仓cash：非保证金
        ordType           	String  	子订单类型market：市价单
        sz                	String  	子订单委托数量
        state             	String  	子订单状态canceled：撤单成功live：等待成交partially_filled：部分成交filled：完全成交cancelling：撤单中
        side              	String  	子订单订单方向buy：买sell：卖
        px                	String  	子订单委托价格市价委托时为"-1"
        fee               	String  	子订单手续费数量
        feeCcy            	String  	子订单手续费币种
        avgPx             	String  	子订单平均成交价格
        accFillSz         	String  	子订单累计成交数量
        tag               	String  	订单标签
        '''
        return self.send_request(*_RecurringBuyEndpoints.get_sub_orders, **to_local(locals()))



from okx.api import Account as AccountAPI
from okx.app import exception
from copy import deepcopy


class AccountSPOT():
    instType = 'SPOT'

    def __init__(
            self,
            key: str,
            secret: str,
            passphrase: str,
            proxies={},
            proxy_host: str = None,
    ):
        FLAG = '0'
        self.api = AccountAPI(
            key=key,
            secret=secret,
            passphrase=passphrase,
            flag=FLAG,
            proxies=proxies,
            proxy_host=proxy_host,
        )

    # 查看账户余额 details为列表
    def get_balances(self, instIds=[], ccys=[]):
        '''
        https://www.okx.com/docs-v5/zh/#rest-api-account-get-balance
        '''
        if instIds or ccys:
            inner_ccys = deepcopy(ccys)
            for instId in instIds:
                _ccy = instId.split('-')[0]
                if _ccy not in inner_ccys:
                    inner_ccys.append(_ccy)
            request_ccy = ','.join(inner_ccys)
            result = self.api.get_balance(ccy=request_ccy)
        else:
            result = self.api.get_balance()
        result['data'] = result['data'][0]
        return result

    # 查看账户余额 details为字典
    def get_balancesMap(self, instIds=[], ccys=[]):
        '''
        https://www.okx.com/docs-v5/zh/#rest-api-account-get-balance
        '''
        result = self.get_balances(instIds=instIds, ccys=ccys)
        data_map = {}
        for data in result['data']['details']:
            ccy = data['ccy']
            data_map[ccy] = data
        result['data']['details'] = data_map
        return result

    # 获取单个币种的余额
    def get_balance(self, instId: str = '', ccy: str = ''):
        '''
        https://www.okx.com/docs-v5/zh/#rest-api-account-get-balance

        :param instId: 产品ID
        :param ccy: 币种
        注意：
            1. instId 与 ccy 必须填写一个
            2. 当instId和ccy同时存在 优先级：ccy > instId
        '''
        if not ccy and not instId:
            msg = 'ccy and instId can not be empty together.'
            raise exception.ParamException(msg)
        if not ccy:
            ccy = instId.split('-')[0]
        result = self.api.get_balance(ccy=ccy)
        if not result['code'] == '0':
            return result
        result['data'] = result['data'][0]
        if len(result['data']['details']) == 0:
            result['data']['details'] = {}
        elif len(result['data']['details']) == 1:
            result['data']['details'] = result['data']['details'][0]
        else:
            msg = 'details got multi data,use get_balances instead.'
            raise exception.UnexpectedException(msg)
        return result



'''
公共数据
https://www.okx.com/docs-v5/zh/#public-data
'''

from paux.param import to_local
from okx.api._client import Client


class _PublicEndpoints():
    get_instruments = ['/api/v5/public/instruments', 'GET']  # 获取交易产品基础信息
    get_delivery_exercise_history = ['/api/v5/public/delivery-exercise-history', 'GET']  # 获取交割和行权记录
    get_open_interest = ['/api/v5/public/open-interest', 'GET']  # 获取持仓总量
    get_funding_rate = ['/api/v5/public/funding-rate', 'GET']  # 获取永续合约当前资金费率
    get_funding_rate_history = ['/api/v5/public/funding-rate-history', 'GET']  # 获取永续合约历史资金费率
    get_price_limit = ['/api/v5/public/price-limit', 'GET']  # 获取限价
    get_opt_summary = ['/api/v5/public/opt-summary', 'GET']  # 获取期权定价
    get_estimated_price = ['/api/v5/public/estimated-price', 'GET']  # 获取预估交割/行权价格
    get_discount_rate_interest_free_quota = ['/api/v5/public/discount-rate-interest-free-quota', 'GET']  # 获取免息额度和币种折算率等级
    get_time = ['/api/v5/public/time', 'GET']  # 获取系统时间
    get_mark_price = ['/api/v5/public/mark-price', 'GET']  # 获取标记价格
    get_position_tiers = ['/api/v5/public/position-tiers', 'GET']  # 获取衍生品仓位档位
    get_interest_rate_loan_quota = ['/api/v5/public/interest-rate-loan-quota', 'GET']  # 获取市场借币杠杆利率和借币限额
    get_vip_interest_rate_loan_quota = ['/api/v5/public/vip-interest-rate-loan-quota', 'GET']  # 获取尊享借币杠杆利率和借币限额
    get_underlying = ['/api/v5/public/underlying', 'GET']  # 获取衍生品标的指数
    get_insurance_fund = ['/api/v5/public/insurance-fund', 'GET']  # 获取风险准备金余额
    get_convert_contract_coin = ['/api/v5/public/convert-contract-coin', 'GET']  # 张币转换
    get_instrument_tick_bands = ['/api/v5/public/instrument-tick-bands', 'GET']  # 获取期权价格梯度
    get_index_tickers = ['/api/v5/market/index-tickers', 'GET']  # 获取指数行情
    get_index_candles = ['/api/v5/market/index-candles', 'GET']  # 获取指数K线数据
    get_history_index_candles = ['/api/v5/market/history-index-candles', 'GET']  # 获取指数历史K线数据
    get_mark_price_candles = ['/api/v5/market/mark-price-candles', 'GET']  # 获取标记价格K线数据
    get_history_mark_price_candles = ['/api/v5/market/history-mark-price-candles', 'GET']  # 获取标记价格历史K线数据
    get_open_oracle = ['/api/v5/market/open-oracle', 'GET']  # Oracle  上链交易数据
    get_exchange_rate = ['/api/v5/market/exchange-rate', 'GET']  # 获取法币汇率
    get_index_components = ['/api/v5/market/index-components', 'GET']  # 获取指数成分数据


class Public(Client):

    # 获取交易产品基础信息
    def get_instruments(self, instType: str, uly: str = '', instFamily: str = '', instId: str = '', proxies={},
                        proxy_host: str = None):
        '''
        获取所有可交易产品的信息列表。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-instruments
        
        限速：20次/2s
        限速规则：IP +instType
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型SPOT：币币MARGIN：币币杠杆SWAP：永续合约FUTURES：交割合约OPTION：期权
        uly               	String  	可选      	标的指数，仅适用于交割/永续/期权，期权必填
        instFamily        	String  	否       	交易品种，仅适用于交割/永续/期权
        instId            	String  	否       	产品ID
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型
        instId            	String  	产品id， 如BTC-USD-SWAP
        uly               	String  	标的指数，如BTC-USD，仅适用于交割/永续/期权
        instFamily        	String  	交易品种，如BTC-USD，仅适用于交割/永续/期权
        category          	String  	币种类别（已废弃）
        baseCcy           	String  	交易货币币种，如BTC-USDT中的BTC，仅适用于币币/币币杠杆
        quoteCcy          	String  	计价货币币种，如BTC-USDT中的USDT，仅适用于币币/币币杠杆
        settleCcy         	String  	盈亏结算和保证金币种，如BTC仅适用于交割/永续/期权
        ctVal             	String  	合约面值，仅适用于交割/永续/期权
        ctMult            	String  	合约乘数，仅适用于交割/永续/期权
        ctValCcy          	String  	合约面值计价币种，仅适用于交割/永续/期权
        optType           	String  	期权类型，C或P仅适用于期权
        stk               	String  	行权价格，仅适用于期权
        listTime          	String  	上线时间Unix时间戳的毫秒数格式，如1597026383085
        expTime           	String  	产品下线时间适用于币币/杠杆/交割/永续/期权，对于交割/期权，为交割/行权日期；亦可以为产品下线时间，有变动就会推送。
        lever             	String  	该instId支持的最大杠杆倍数，不适用于币币、期权
        tickSz            	String  	下单价格精度，如0.0001对于期权来说，是梯度中的最小下单价格精度，如果想要获取期权价格梯度，请使用"获取期权价格梯度"接口
        lotSz             	String  	下单数量精度，如 BTC-USDT-SWAP：1
        minSz             	String  	最小下单数量,合约的数量单位是“张”，现货的数量单位是“交易货币”
        ctType            	String  	linear：正向合约inverse：反向合约仅适用于交割/永续
        alias             	String  	合约日期别名this_week：本周next_week：次周quarter：季度next_quarter：次季度仅适用于交割
        state             	String  	产品状态live：交易中suspend：暂停中preopen：预上线，如：交割和期权的新合约在 live 之前，会有 preopen 状态test：测试中（测试产品，不可交易）
        maxLmtSz          	String  	合约或现货限价单的单笔最大委托数量,合约的数量单位是“张”，现货的数量单位是“交易货币”
        maxMktSz          	String  	合约或现货市价单的单笔最大委托数量,合约的数量单位是“张”，现货的数量单位是“USDT”
        maxTwapSz         	String  	合约或现货时间加权单的单笔最大委托数量,合约的数量单位是“张”，现货的数量单位是“交易货币”
        maxIcebergSz      	String  	合约或现货冰山委托的单笔最大委托数量,合约的数量单位是“张”，现货的数量单位是“交易货币”
        maxTriggerSz      	String  	合约或现货计划委托委托的单笔最大委托数量,合约的数量单位是“张”，现货的数量单位是“交易货币”
        maxStopSz         	String  	合约或现货止盈止损市价委托的单笔最大委托数量,合约的数量单位是“张”，现货的数量单位是“USDT”
        '''
        return self.send_request(*_PublicEndpoints.get_instruments, **to_local(locals()))

    # 获取交割和行权记录
    def get_delivery_exercise_history(self, instType: str, uly: str = '', instFamily: str = '', after: str = '',
                                      before: str = '', limit: str = '', proxies={}, proxy_host: str = None):
        '''
        获取3个月内的交割合约的交割记录和期权的行权记录
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-delivery-exercise-history
        
        限速：40次/2s
        限速规则：IP + (instrumentType + uly)
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型FUTURES：交割合约OPTION：期权
        uly               	String  	可选      	标的指数uly与instFamily必须传一个,若传两个，以instFamily为主
        instFamily        	String  	可选      	交易品种uly与instFamily必须传一个,若传两个，以instFamily为主
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	交割/行权日期，Unix时间戳的毫秒数格式，如1597026383085
        details           	String  	详细数据
        > instId          	String  	交割/行权的合约ID
        > px              	String  	交割/行权的价格
        > type            	String  	类型delivery：交割exercised：实值已行权expired_otm：虚值已过期
        '''
        return self.send_request(*_PublicEndpoints.get_delivery_exercise_history, **to_local(locals()))

    # 获取持仓总量
    def get_open_interest(self, instType: str, uly: str = '', instFamily: str = '', instId: str = '', proxies={},
                          proxy_host: str = None):
        '''
        查询单个交易产品的市场的持仓总量
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-open-interest
        
        限速：20次/2s
        限速规则：IP + instrumentID
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型SWAP：永续合约FUTURES：交割合约OPTION：期权
        uly               	String  	可选      	标的指数适用于交割/永续/期权期权情况下，uly和instFamily必须传一个
        instFamily        	String  	可选      	交易品种适用于交割/永续/期权期权情况下，uly和instFamily必须传一个
        instId            	String  	否       	产品ID，如BTC-USD-180216仅适用于交割/永续/期权
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型
        instId            	String  	产品ID
        oi                	String  	持仓量（按张折算）
        oiCcy             	String  	持仓量（按币折算）
        ts                	String  	数据返回时间，Unix时间戳的毫秒数格式 ，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_open_interest, **to_local(locals()))

    # 获取永续合约当前资金费率
    def get_funding_rate(self, instId: str, proxies={}, proxy_host: str = None):
        '''
        获取当前资金费率
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-funding-rate
        
        限速：20次/2s
        限速规则：IP +instrumentID
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID ，如BTC-USD-SWAP仅适用于永续
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型SWAP：永续合约
        instId            	String  	产品ID，如BTC-USD-SWAP
        fundingRate       	String  	资金费率
        nextFundingRate   	String  	下一期预测资金费率
        fundingTime       	String  	资金费时间 ，Unix时间戳的毫秒数格式，如1597026383085
        nextFundingTime   	String  	下一期资金费时间 ，Unix时间戳的毫秒数格式，如1622851200000
        '''
        return self.send_request(*_PublicEndpoints.get_funding_rate, **to_local(locals()))

    # 获取永续合约历史资金费率
    def get_funding_rate_history(self, instId: str, before: str = '', after: str = '', limit: str = '', proxies={},
                                 proxy_host: str = None):
        '''
        获取最近3个月的历史资金费率
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-funding-rate-history
        
        限速：10次/2s
        限速规则：IP +instrumentID
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID ，如BTC-USD-SWAP仅适用于永续
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的fundingTime
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的fundingTime
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型SWAP：永续合约
        instId            	String  	产品ID，如BTC-USD-SWAP
        fundingRate       	String  	预计资金费率
        realizedRate      	String  	实际资金费率
        fundingTime       	String  	资金费时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_funding_rate_history, **to_local(locals()))

    # 获取限价
    def get_price_limit(self, instId: str, proxies={}, proxy_host: str = None):
        '''
        查询单个交易产品的最高买价和最低卖价
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-limit-price
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USDT-SWAP仅适用于交割/永续/期权
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型SWAP：永续合约FUTURES：交割合约OPTION：期权
        instId            	String  	产品ID ，如BTC-USDT-SWAP
        buyLmt            	String  	最高买价
        sellLmt           	String  	最低卖价
        ts                	String  	限价数据更新时间 ，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_price_limit, **to_local(locals()))

    # 获取期权定价
    def get_opt_summary(self, uly: str = '', instFamily: str = '', expTime: str = '', proxies={},
                        proxy_host: str = None):
        '''
        查询期权详细信息
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-option-market-data
        
        限速：20次/2s
        限速规则：IP +uly
    
        请求参数:
        Parameter         	Type    	Required	Description

        uly               	String  	可选      	标的指数，仅适用于期权uly与instFamily必须传一个,若传两个，以instFamily为主
        instFamily        	String  	可选      	交易品种，仅适用于期权uly与instFamily必须传一个,若传两个，以instFamily为主
        expTime           	String  	否       	合约到期日，格式为"YYMMDD"，如 "200527"
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型OPTION：期权
        instId            	String  	产品ID，如BTC-USD-200103-5500-C
        uly               	String  	标的指数
        delta             	String  	期权价格对uly价格的敏感度
        gamma             	String  	delta对uly价格的敏感度
        vega              	String  	权价格对隐含波动率的敏感度
        theta             	String  	期权价格对剩余期限的敏感度
        deltaBS           	String  	BS模式下期权价格对uly价格的敏感度
        gammaBS           	String  	BS模式下delta对uly价格的敏感度
        vegaBS            	String  	BS模式下期权价格对隐含波动率的敏感度
        thetaBS           	String  	BS模式下期权价格对剩余期限的敏感度
        lever             	String  	杠杆倍数
        markVol           	String  	标记波动率
        bidVol            	String  	bid波动率
        askVol            	String  	ask波动率
        realVol           	String  	已实现波动率（目前该字段暂未启用）
        volLv             	String  	价平期权的隐含波动率
        fwdPx             	String  	远期价格
        ts                	String  	数据更新时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_opt_summary, **to_local(locals()))

    # 获取预估交割/行权价格
    def get_estimated_price(self, instId: str, proxies={}, proxy_host: str = None):
        '''
        获取交割合约和期权预估交割/行权价。交割/行权预估价只有交割/行权前一小时才有返回值
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-estimated-delivery-exercise-price
        
        限速：10次/2s
        限速规则：IP +instrumentID
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID， 如BTC-USD-200214仅适用于交割/期权
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型FUTURES：交割合约OPTION：期权
        instId            	String  	产品ID， 如BTC-USD-180216
        settlePx          	String  	预估交割、行权价格
        ts                	String  	数据返回时间 ，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_estimated_price, **to_local(locals()))

    # 获取免息额度和币种折算率等级
    def get_discount_rate_interest_free_quota(self, ccy: str = '', discountLv: str = '', proxies={},
                                              proxy_host: str = None):
        '''
        获取免息额度和币种折算率等级
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-discount-rate-and-interest-free-quota
        
        限速：2 次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        ccy               	String  	否       	币种
        discountLv        	String  	否       	折算率等级1:第一档2:第二档3:第三档4:第四档5:第五档
        返回参数:
        Parameter         	Type    	Description
        ccy               	String  	币种
        amt               	String  	免息金额
        discountLv        	String  	折算率等级折算率等级说明
        discountInfo      	Array   	币种折算率详情
        > discountRate    	String  	折算率
        > maxAmt          	String  	梯度区间上限（美元）， “”表示正无穷
        > minAmt          	String  	梯度区间下限（美元）， 最小值是0
        '''
        return self.send_request(*_PublicEndpoints.get_discount_rate_interest_free_quota, **to_local(locals()))

    # 获取系统时间
    def get_time(self, proxies={}, proxy_host: str = None):
        '''
        获取系统时间
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-system-time
        
        限速：10次/2s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	系统时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_time, **to_local(locals()))

    # 获取标记价格
    def get_mark_price(self, instType: str, uly: str = '', instFamily: str = '', instId: str = '', proxies={},
                       proxy_host: str = None):
        '''
        为了防止个别用户恶意操控市场导致合约价格波动剧烈，我们根据现货指数和合理基差设定标记价格。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-mark-price
        
        限速：10次/2s
        限速规则：IP +instrumentID
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型MARGIN：币币杠杆SWAP：永续合约FUTURES：交割合约OPTION：期权
        uly               	String  	否       	标的指数适用于交割/永续/期权
        instFamily        	String  	否       	交易品种适用于交割/永续/期权
        instId            	String  	否       	产品ID，如BTC-USD-SWAP
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型MARGIN：币币杠杆SWAP：永续合约FUTURES：交割合约OPTION：期权
        instId            	String  	产品ID，如BTC-USD-200214
        markPx            	String  	标记价格
        ts                	String  	接口数据返回时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_mark_price, **to_local(locals()))

    # 获取衍生品仓位档位
    def get_position_tiers(self, instType: str, tdMode: str, uly: str = '', instFamily: str = '', instId: str = '',
                           ccy: str = '', tier: str = '', proxies={}, proxy_host: str = None):
        '''
        全部仓位档位对应信息，当前最高可开杠杆倍数由您的借币持仓和保证金率决定。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-position-tiers
        
        限速：10次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型MARGIN：币币杠杆SWAP：永续合约FUTURES：交割合约OPTION：期权
        tdMode            	String  	是       	保证金模式isolated：逐仓 ；cross：全仓
        uly               	String  	可选      	标的指数，支持多uly，半角逗号分隔，最大不超过3个当产品类型是永续、交割、期权之一时，uly与instFamily必须传一个，若传两个，以instFamily为主当产品类型是MARGIN时忽略
        instFamily        	String  	可选      	交易品种，支持多instFamily，半角逗号分隔，最大不超过5个当产品类型是永续、交割、期权之一时，uly与instFamily必须传一个，若传两个，以instFamily为主
        instId            	String  	可选      	产品ID，支持多instId，半角逗号分隔，最大不超过5个仅适用币币杠杆，instId和ccy必须传一个，若传两个，以instId为主
        ccy               	String  	可选      	保证金币种仅适用杠杆全仓，该值生效时，返回的是跨币种保证金模式和组合保证金模式下的借币量
        tier              	String  	否       	查指定档位
        返回参数:
        Parameter         	Type    	Description
        uly               	String  	标的指数适用于交割/永续/期权
        instFamily        	String  	交易品种适用于交割/永续/期权
        instId            	String  	币对
        tier              	String  	仓位档位
        minSz             	String  	该档位最少借币量或者持仓数量  杠杆/期权/永续/交割 最小持仓量 默认0当ccy参数生效时，返回ccy的最小借币量
        maxSz             	String  	该档位最多借币量或者持仓数量  杠杆/期权/永续/交割当ccy参数生效时，返回ccy的最大借币量
        mmr               	String  	维持保证金率
        imr               	String  	最低初始保证金率
        maxLever          	String  	最高可用杠杆倍数
        optMgnFactor      	String  	期权保证金系数  （仅适用于期权）
        quoteMaxLoan      	String  	计价货币  最大借币量（仅适用于杠杆，且instId参数生效时），例如 BTC-USDT   里的 USDT最大借币量
        baseMaxLoan       	String  	交易货币  最大借币量（仅适用于杠杆，且instId参数生效时），例如 BTC-USDT   里的 BTC最大借币量
        '''
        return self.send_request(*_PublicEndpoints.get_position_tiers, **to_local(locals()))

    # 获取市场借币杠杆利率和借币限额
    def get_interest_rate_loan_quota(self, proxies={}, proxy_host: str = None):
        '''
        GET /api/v5/public/interest-rate-loan-quota
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-interest-rate-and-loan-quota
        
        限速：2次/2s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        basic             	Array   	基础利率和借币限额
        > ccy             	String  	币种
        > rate            	String  	基础杠杆日利率
        > quota           	String  	基础借币限额
        vip               	Array   	专业用户
        > level           	String  	账户交易手续费等级，如VIP1
        > loanQuotaCoef   	String  	借币限额系数
        > irDiscount      	String  	利率的折扣率(已废弃)
        regular           	Array   	普通用户
        > level           	String  	账户交易手续费等级，如Lv1
        > loanQuotaCoef   	String  	借币限额系数
        > irDiscount      	String  	利率的折扣率(已废弃)
        '''
        return self.send_request(*_PublicEndpoints.get_interest_rate_loan_quota, **to_local(locals()))

    # 获取尊享借币杠杆利率和借币限额
    def get_vip_interest_rate_loan_quota(self, proxies={}, proxy_host: str = None):
        '''
        GET /api/v5/public/vip-interest-rate-loan-quota
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-interest-rate-and-loan-quota-for-vip-loans
        
        限速：2次/2s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        ccy               	String  	币种
        rate              	String  	借币日利率
        quota             	String  	基础借币限额(已废弃)
        levelList         	Array   	不同VIP等级下的借币限额信息
        > level           	String  	用户VIP等级，如VIP5
        > loanQuota       	String  	借币限额
        '''
        return self.send_request(*_PublicEndpoints.get_vip_interest_rate_loan_quota, **to_local(locals()))

    # 获取衍生品标的指数
    def get_underlying(self, instType: str, proxies={}, proxy_host: str = None):
        '''
        GET /api/v5/public/underlying
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-underlying
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型SWAP：永续合约FUTURES：交割合约OPTION：期权
        返回参数:
        Parameter         	Type    	Description
        uly               	Array   	标的指数   如：BTC-USDT
        '''
        return self.send_request(*_PublicEndpoints.get_underlying, **to_local(locals()))

    # 获取风险准备金余额
    def get_insurance_fund(self, instType: str, type: str = '', uly: str = '', instFamily: str = '', ccy: str = '',
                           before: str = '', after: str = '', limit: str = '', proxies={}, proxy_host: str = None):
        '''
        通过该接口获取系统风险准备金余额信息
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-insurance-fund
        
        限速：10次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	是       	产品类型MARGIN：币币杠杆SWAP：永续合约FUTURES：交割合约OPTION：期权
        type              	String  	否       	风险准备金类型liquidation_balance_deposit：强平注入 ；bankruptcy_loss：穿仓亏损 ；platform_revenue：平台收入注入默认返回全部类型
        uly               	String  	可选      	标的指数交割/永续/期权情况下，uly与instFamily必须传一个，若传两个，以instFamily为主
        instFamily        	String  	可选      	交易品种交割/永续/期权情况下，uly与instFamily必须传一个，若传两个，以instFamily为主
        ccy               	String  	可选      	币种， 仅适用币币杠杆，且必填写
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        total             	String  	平台风险准备金总计，单位为USD
        instFamily        	String  	交易品种适用于交割/永续/期权
        details           	Array of objects	风险准备金详情
        > balance         	String  	风险准备金总量
        > amt             	String  	风险准备金更新数量
        > ccy             	String  	风险准备金总量对应的币种
        > type            	String  	风险准备金类型
        > ts              	String  	风险准备金更新时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_insurance_fund, **to_local(locals()))

    # 张币转换
    def get_convert_contract_coin(self, instId: str, sz: str, type: str = '', px: str = '', unit: str = '',
                                  opType: str = '', proxies={}, proxy_host: str = None):
        '''
        由币转换为张，或者张转换为币。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-unit-convert
        
        限速：10次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        type              	String  	否       	转换类型1: 币转张2: 张转币默认为 1
        instId            	String  	是       	产品ID，仅适用于交割/永续/期权
        sz                	String  	是       	数量，币转张时，为币的数量，张转币时，为张的数量。张的数量，只能是正整数
        px                	String  	可选      	委托价格币本位合约的张币转换时必填；U本位合约，usdt 与张的转换时，必填；coin 与张的转换时，可不填；期权的张币转换时，可不填。
        unit              	String  	否       	币的单位，coin: 币，usds: usdt 或者 usdc仅适用于交割和永续合约的U本位合约
        opType            	String  	否       	将要下单的类型open: 开仓时将sz舍位close: 平仓时将sz四舍五入默认值为close适用于交割永续
        返回参数:
        Parameter         	Type    	Description
        type              	String  	转换类型1: 币转张2: 张转币
        instId            	String  	产品ID
        px                	String  	委托价格
        sz                	String  	数量，张转币时，为币的数量，币转张时，为张的数量
        unit              	String  	币的单位，coin: 币，usds: usdt 或者 usdc
        '''
        return self.send_request(*_PublicEndpoints.get_convert_contract_coin, **to_local(locals()))

    # 获取期权价格梯度
    def get_instrument_tick_bands(self, instType: str = '', instFamily: str = '', proxies={}, proxy_host: str = None):
        '''
        获取期权价格梯度信息
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-option-tickbands
        
        限速：5次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instType          	String  	Yes     	产品类型OPTION：期权
        instFamily        	String  	No      	交易品种，仅适用于期权
        返回参数:
        Parameter         	Type    	Description
        instType          	String  	产品类型
        instFamily        	String  	交易品种
        tickBand          	String  	下单价格精度梯度
        > minPx           	String  	最低下单价格
        > maxPx           	String  	最高下单价格
        > tickSz          	String  	下单价格精度，如 0.0001
        '''
        return self.send_request(*_PublicEndpoints.get_instrument_tick_bands, **to_local(locals()))

    # 获取指数行情
    def get_index_tickers(self, quoteCcy: str = '', instId: str = '', proxies={}, proxy_host: str = None):
        '''
        获取指数行情数据
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-index-tickers
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        quoteCcy          	String  	可选      	指数计价单位， 目前只有USD/USDT/BTC/USDC为计价单位的指数，quoteCcy和instId必须填写一个
        instId            	String  	可选      	指数，如BTC-USD
        返回参数:
        Parameter         	Type    	Description
        instId            	String  	指数
        idxPx             	String  	最新指数价格
        high24h           	String  	24小时指数最高价格
        low24h            	String  	24小时指数最低价格
        open24h           	String  	24小时指数开盘价格
        sodUtc0           	String  	UTC 0  时开盘价
        sodUtc8           	String  	UTC+8 时开盘价
        ts                	String  	指数价格更新时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_PublicEndpoints.get_index_tickers, **to_local(locals()))

    # 获取指数K线数据
    def get_index_candles(self, instId: str, after: str = '', before: str = '', bar: str = '', limit: str = '',
                          proxies={}, proxy_host: str = None):
        '''
        指数K线数据每个粒度最多可获取最近1,440条。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-index-candlesticks
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	现货指数，如BTC-USD
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts, 单独使用时，会返回最新的数据。
        bar               	String  	否       	时间粒度，默认值1m如 [1m/3m/5m/15m/30m/1H/2H/4H]香港时间开盘价k线：[6H/12H/1D/1W/1M/3M]UTC时间开盘价k线：[6Hutc/12Hutc/1Dutc/1Wutc/1Mutc/3Mutc]
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	开始时间，Unix时间戳的毫秒数格式，如1597026383085
        o                 	String  	开盘价格
        h                 	String  	最高价格
        l                 	String  	最低价格
        c                 	String  	收盘价格
        confirm           	String  	K线状态0代表 K 线未完结，1代表 K 线已完结。
        '''
        return self.send_request(*_PublicEndpoints.get_index_candles, **to_local(locals()))

    # 获取指数历史K线数据
    def get_history_index_candles(self, instId: str, after: str = '', before: str = '', bar: str = '', limit: str = '',
                                  proxies={}, proxy_host: str = None):
        '''
        获取最近几年的指数K线数据
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-index-candlesticks-history
        
        限速：10次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	现货指数，如BTC-USD
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts, 单独使用时，会返回最新的数据。
        bar               	String  	否       	时间粒度，默认值1m如 [1m/3m/5m/15m/30m/1H/2H/4H]香港时间开盘价k线：[6H/12H/1D/1W/1M]UTC时间开盘价k线：[/6Hutc/12Hutc/1Dutc/1Wutc/1Mutc]
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	开始时间，Unix时间戳的毫秒数格式，如1597026383085
        o                 	String  	开盘价格
        h                 	String  	最高价格
        l                 	String  	最低价格
        c                 	String  	收盘价格
        confirm           	String  	K线状态0代表 K 线未完结，1代表 K 线已完结。
        '''
        return self.send_request(*_PublicEndpoints.get_history_index_candles, **to_local(locals()))

    # 获取标记价格K线数据
    def get_mark_price_candles(self, instId: str, after: str = '', before: str = '', bar: str = '', limit: str = '',
                               proxies={}, proxy_host: str = None):
        '''
        标记价格K线数据每个粒度最多可获取最近1,440条。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-mark-price-candlesticks
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USD-SWAP
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts, 单独使用时，会返回最新的数据。
        bar               	String  	否       	时间粒度，默认值1m如 [1m/3m/5m/15m/30m/1H/2H/4H]香港时间开盘价k线：[6H/12H/1D/1W/1M/3M]UTC时间开盘价k线：[6Hutc/12Hutc/1Dutc/1Wutc/1Mutc/3Mutc]
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	开始时间，Unix时间戳的毫秒数格式，如1597026383085
        o                 	String  	开盘价格
        h                 	String  	最高价格
        l                 	String  	最低价格
        c                 	String  	收盘价格
        confirm           	String  	K线状态0代表 K 线未完结，1代表 K 线已完结。
        '''
        return self.send_request(*_PublicEndpoints.get_mark_price_candles, **to_local(locals()))

    # 获取标记价格历史K线数据
    def get_history_mark_price_candles(self, instId: str, after: str = '', before: str = '', bar: str = '',
                                       limit: str = '', proxies={}, proxy_host: str = None):
        '''
        获取最近几年的标记价格K线数据
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-mark-price-candlesticks-history
        
        限速：10次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        instId            	String  	是       	产品ID，如BTC-USD-SWAP
        after             	String  	否       	请求此时间戳之前（更旧的数据）的分页内容，传的值为对应接口的ts
        before            	String  	否       	请求此时间戳之后（更新的数据）的分页内容，传的值为对应接口的ts, 单独使用时，会返回最新的数据。
        bar               	String  	否       	时间粒度，默认值1m如 [1m/3m/5m/15m/30m/1H/2H/4H]香港时间开盘价k线：[6H/12H/1D/1W/1M]UTC时间开盘价k线：[6Hutc/12Hutc/1Dutc/1Wutc/1Mutc]
        limit             	String  	否       	分页返回的结果集数量，最大为100，不填默认返回100条
        返回参数:
        Parameter         	Type    	Description
        ts                	String  	开始时间，Unix时间戳的毫秒数格式，如1597026383085
        o                 	String  	开盘价格
        h                 	String  	最高价格
        l                 	String  	最低价格
        c                 	String  	收盘价格
        confirm           	String  	K线状态0代表 K 线未完结，1代表 K 线已完结。
        '''
        return self.send_request(*_PublicEndpoints.get_history_mark_price_candles, **to_local(locals()))

    # Oracle  上链交易数据
    def get_open_oracle(self, proxies={}, proxy_host: str = None):
        '''
        获取使用 Open Oracle 智能合约签名后加密资产价格。
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-oracle
        
        限速：1次/5s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        messages          	String  	数组包含对[ kind，timestamp，key，value]进行ABI编码的值，其中kind恒等于prices，timestamp是获取价格的时间戳，key是加密资产（例如，ETH），value是资产价格
        prices            	String  	价格
        signatures        	String  	每个消息的以太坊兼容ECDSA签名的数组
        timestamp         	String  	获取最新数据点的时间，Unix时间戳,如1597026387
        '''
        return self.send_request(*_PublicEndpoints.get_open_oracle, **to_local(locals()))

    # 获取法币汇率
    def get_exchange_rate(self, proxies={}, proxy_host: str = None):
        '''
        该接口提供的是2周的平均汇率数据
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-exchange-rate
        
        限速：1次/2s
        限速规则：IP
    
        
        返回参数:
        Parameter         	Type    	Description
        usdCny            	String  	人民币兑美元汇率
        '''
        return self.send_request(*_PublicEndpoints.get_exchange_rate, **to_local(locals()))

    # 获取指数成分数据
    def get_index_components(self, index: str, proxies={}, proxy_host: str = None):
        '''
        查询市场上的指数成分信息数据
        https://www.okx.com/docs-v5/zh/#public-data-rest-api-get-index-components
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        index             	String  	是       	指数，如BTC-USDT
        返回参数:
        Parameter         	Type    	Description
        index             	String  	指数名称
        last              	String  	最新指数价格
        ts                	String  	数据产生时间，Unix时间戳的毫秒数格式， 如1597026383085
        components        	String  	成分
        > exch            	String  	交易所名称
        > symbol          	String  	采集的币对名称
        > symPx           	String  	采集的币对价格
        > wgt             	String  	权重
        > cnvPx           	String  	换算成指数后的价格
        '''
        return self.send_request(*_PublicEndpoints.get_index_components, **to_local(locals()))



import time
from okx.app import code
from okx.app.market._base import MarketBase


class ExchangeInfo(MarketBase):

    # 以缓存的方式获取全部交易产品基础信息
    def get_exchangeInfos(self, expire_seconds: int = 60 * 5, uly: str = ''):
        '''
        :param expire_seconds: 缓存时间（秒）
        :param uly: 标的指数，仅适用于交割/永续/期权，期权必填
        使用的缓存数据格式：
            self._exchangeInfo_cache = [
                {
                    'code':<状态码>,
                    'data':<exchangeInfo数据>,
                    'msg':<提示信息>,
                },
                <上次更新的毫秒时间戳>
            ]
        '''

        if (
                # 无缓存数据
                not hasattr(self, '_exchangeInfo_caches')
                or
                # 缓存数据过期
                getattr(self, '_exchangeInfo_caches')[1] - time.time() * 1000 >= expire_seconds
        ):
            # 更新数据并设置时间戳
            setattr(self, '_exchangeInfo_caches',
                    [self.publicAPI.get_instruments(instType=self.instType, uly=uly), time.time() * 1000])
        # 返回缓存数据
        return getattr(self, '_exchangeInfo_caches')[0]

    # 以缓存的方式获取单个交易产品基础信息
    def get_exchangeInfo(
            self,
            instId: str,
            expire_seconds: int = 60 * 5,
            uly: str = '',
    ):
        '''
        :param instId: 产品
        :param expire_seconds: 缓存时间（秒）
        :param uly: 标的指数，仅适用于交割/永续/期权，期权必填
        '''
        exchangeInfos_result = self.get_exchangeInfos(uly=uly, expire_seconds=expire_seconds)
        # [ERROR RETURN] 异常交易规则与交易
        if exchangeInfos_result['code'] != '0':
            return exchangeInfos_result
        # 寻找instId的信息
        for instId_data in exchangeInfos_result['data']:
            if instId_data['instId'] == instId:
                instId_data = instId_data
                break
        else:
            instId_data = None
        # [ERROR RETURN] 没有找到instId的交易规则与交易对信息
        if instId_data == None:
            result = {
                'code': code.EXCHANGE_INFO_ERROR[0],
                'data': exchangeInfos_result['data'],
                'msg': f'Symbol not found instId={instId}'
            }
            return result
        # 将filters中的列表转换为字典，里面可能包含下单价格与数量精度
        result = {
            'code': '0',
            'data': instId_data,
            'msg': '',
        }
        return result

    # 获取可以交易的产品列表
    def get_instIds_trading_on(
            self,
            expire_seconds: int = 60 * 5,
            uly: str = '',
    ) -> dict:
        '''
        :param expire_seconds: 缓存时间（秒）
        :param uly: 标的指数，仅适用于交割/永续/期权，期权必填
        '''
        exchangeInfos_result = self.get_exchangeInfos(uly=uly, expire_seconds=expire_seconds)
        # [ERROR RETURN] 异常交易规则与交易
        if exchangeInfos_result['code'] != '0':
            return exchangeInfos_result
        status_name = 'state'

        instIds = [
            data['instId']
            for data in exchangeInfos_result['data']
            if data[status_name] == 'live'
        ]
        # [RETURN]
        result = {
            'code': '0',
            'data': instIds,
            'msg': ''
        }
        return result

    # 获取不可交易的产品列表
    def get_instIds_trading_off(
            self,
            expire_seconds: int = 60 * 5,
            uly: str = '',
    ) -> dict:
        '''
        :param expire_seconds: 缓存时间（秒）
        :param uly: 标的指数，仅适用于交割/永续/期权，期权必填
        '''
        exchangeInfos_result = self.get_exchangeInfos(uly=uly, expire_seconds=expire_seconds)
        # [ERROR RETURN] 异常交易规则与交易
        if exchangeInfos_result['code'] != '0':
            return exchangeInfos_result
        status_name = 'state'

        instIds = [
            data['instId']
            for data in exchangeInfos_result['data']
            if data[status_name] != 'live'
        ]
        # [RETURN]
        result = {
            'code': '0',
            'data': instIds,
            'msg': ''
        }
        return result

    # 获取可以交易的产品列表
    def get_instIds_all(
            self,
            expire_seconds: int = 60 * 5,
            uly: str = '',
    ) -> dict:
        '''
        :param expire_seconds: 缓存时间（秒）
        :param uly: 标的指数，仅适用于交割/永续/期权，期权必填
        '''
        exchangeInfos_result = self.get_exchangeInfos(uly=uly, expire_seconds=expire_seconds)
        # [ERROR RETURN] 异常交易规则与交易
        if exchangeInfos_result['code'] != '0':
            return exchangeInfos_result
        instIds = [
            data['instId'] for data in exchangeInfos_result['data']
        ]
        # [RETURN]
        result = {
            'code': '0',
            'data': instIds,
            'msg': ''
        }
        return result



'''
价差交易
https://www.okx.com/docs-v5/zh/#spread-trading
'''
from paux.param import to_local
from okx.api._client import Client


class _SpreadTradeEndpoints():
    set_order = ['/api/v5/sprd/order', 'POST']  # 下单
    set_cancel_order = ['/api/v5/sprd/cancel-order', 'POST']  # 撤单
    set_mass_cancel = ['/api/v5/sprd/mass-cancel', 'POST']  # 全部撤单
    get_order = ['/api/v5/sprd/order', 'GET']  # 获取订单信息
    get_orders_pending = ['/api/v5/sprd/orders-pending', 'GET']  # 获取未成交订单列表
    get_orders_history = ['/api/v5/sprd/orders-history', 'GET']  # 获取历史订单记录（近七天)
    get_trades = ['/api/v5/sprd/trades', 'GET']  # 获取历史成交数据（近七天）
    get_spreads = ['/api/v5/sprd/spreads', 'GET']  # 获取Spreads（公共）
    get_books = ['/api/v5/sprd/books', 'GET']  # 获取Spread产品深度（公共）
    get_ticker = ['/api/v5/sprd/ticker', 'GET']  # 获取单个Spread产品行情信息（公共）
    get_public_trades = ['/api/v5/sprd/public-trades', 'GET']  # 获取公共成交数据（公共）


class SpreadTrade(Client):

    # 下单
    def set_order(self, sprdId: str, side: str, ordType: str, sz: str, px: str, clOrdId: str = '', tag: str = '',
                  proxies={}, proxy_host: str = None):
        '''
        下单
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-place-order
        
        
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	是       	spread ID，如 BTC-USDT_BTC-USDT-SWAP
        clOrdId           	String  	否       	客户自定义订单ID字母（区分大小写）与数字的组合，可以是纯字母、纯数字且长度要在1-32位之间。
        tag               	String  	否       	订单标签字母（区分大小写）与数字的组合，可以是纯字母、纯数字，且长度在1-16位之间。
        side              	String  	是       	订单方向buy：买，sell：卖
        ordType           	String  	是       	订单类型limit：限价单post_only：只做maker单
        sz                	String  	是       	委托数量。反向价差的数量单位为USD，正向价差为其对应baseCcy
        px                	String  	是       	委托价格，仅适用于limit,post_only类型的订单
        返回参数:
        Parameter         	Type    	Description
        ordId             	String  	订单ID
        clOrdId           	String  	客户自定义订单ID
        tag               	String  	订单标签
        sCode             	String  	事件执行结果的code，0代表成功
        sMsg              	String  	事件执行失败或成功时的msg
        '''
        return self.send_request(*_SpreadTradeEndpoints.set_order, **to_local(locals()))

    # 撤单
    def set_cancel_order(self, ordId: str = '', clOrdId: str = '', proxies={}, proxy_host: str = None):
        '''
        撤销之前下的未完成订单。
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-cancel-order
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        ordId             	String  	可选      	订单ID，ordId和clOrdId必须传一个，若传两个，以ordId为主
        clOrdId           	String  	可选      	用户自定义ID
        返回参数:
        Parameter         	Type    	Description
        ordId             	String  	订单ID
        clOrdId           	String  	客户自定义订单ID
        sCode             	String  	事件执行结果的code，0代表成功
        sMsg              	String  	事件执行失败时的msg
        '''
        return self.send_request(*_SpreadTradeEndpoints.set_cancel_order, **to_local(locals()))

    # 全部撤单
    def set_mass_cancel(self, sprdId: str = '', proxies={}, proxy_host: str = None):
        '''
        撤销所有挂单
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-cancel-all-orders
        
        限速：10次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	否       	spread ID
        返回参数:
        Parameter         	Type    	Description
        result            	Boolean 	请求结果true,false
        '''
        return self.send_request(*_SpreadTradeEndpoints.set_mass_cancel, **to_local(locals()))

    # 获取订单信息
    def get_order(self, ordId: str = '', clOrdId: str = '', proxies={}, proxy_host: str = None):
        '''
        查订单信息
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-order-details
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        ordId             	String  	可选      	订单ID，ordId和clOrdId必须传一个，若传两个，以ordId为主
        clOrdId           	String  	可选      	用户自定义ID，如果clOrdId关联了多个订单，只会返回最近的那笔订单
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	Spread ID
        ordId             	String  	订单ID
        clOrdId           	String  	客户自定义订单ID
        tag               	String  	订单标签
        px                	String  	委托价格
        sz                	String  	委托数量
        ordType           	String  	订单类型limit：限价单
        side              	String  	订单方向
        fillSz            	String  	最新成交数量
        fillPx            	String  	最新成交价格
        tradeId           	String  	最近成交ID
        accFillSz         	String  	累计成交数量
        pendingFillSz     	String  	待成交数量（包括带结算数量）
        pendingSettleSz   	String  	待结算数量
        canceledSz        	String  	被取消数量
        avgPx             	String  	成交均价，如果成交数量为0，该字段为""
        state             	String  	订单状态canceled：撤单成功live：等待成交partially_filled：部分成交filled：完全成交
        cancelSource      	String  	撤单原因0: 系统撤单1: 用户撤单31: 当前只挂单订单 (Post only) 将会吃掉挂单深度32: 自成交保护撤单
        uTime             	String  	订单状态更新时间，Unix时间戳的毫秒数格式，如1597026383085
        cTime             	String  	订单创建时间，Unix时间戳的毫秒数格式， 如1597026383085
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_order, **to_local(locals()))

    # 获取未成交订单列表
    def get_orders_pending(self, sprdId: str = '', ordType: str = '', state: str = '', beginId: str = '',
                           endId: str = '', limit: str = '', proxies={}, proxy_host: str = None):
        '''
        获取当前账户下所有未成交订单信息
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-active-orders
        
        限速：10次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	否       	spread ID，如BTC-USDT_BTC-USDT-SWAP
        ordType           	String  	否       	订单类型limit：限价单
        state             	String  	否       	订单状态live：等待成交partially_filled：部分成交
        beginId           	String  	否       	请求的起始订单ID，请求此ID之后（更新的数据）的分页内容，不包括 beginId
        endId             	String  	否       	请求的结束订单ID，请求此ID之前（更旧的数据）的分页内容，不包括 endId
        limit             	String  	否       	返回结果的数量，最大为100，默认100条
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	spread ID，如BTC-USDT_BTC-USDT-SWAP
        ordId             	String  	订单ID
        clOrdId           	String  	客户自定义订单ID
        tag               	String  	订单标签
        px                	String  	委托价格
        sz                	String  	委托数量
        ordType           	String  	订单类型limit：限价单
        side              	String  	订单方向
        fillSz            	String  	最新成交数量
        fillPx            	String  	最新成交价格
        tradeId           	String  	最近成交ID
        accFillSz         	String  	累计成交数量
        pendingFillSz     	String  	待成交数量（包括带结算数量）
        pendingSettleSz   	String  	待结算数量
        canceledSz        	String  	被取消数量
        avgPx             	String  	成交均价，如果成交数量为0，该字段为""
        state             	String  	订单状态live：等待成交partially_filled：部分成交
        cancelSource      	String  	撤单原因0: 系统撤单1: 用户撤单31: 当前只挂单订单 (Post only) 将会吃掉挂单深度
        uTime             	String  	订单状态更新时间，Unix时间戳的毫秒数格式，如：1597026383085
        cTime             	String  	订单创建时间，Unix时间戳的毫秒数格式，如：1597026383085
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_orders_pending, **to_local(locals()))

    # 获取历史订单记录（近七天)
    def get_orders_history(self, sprdId: str = '', ordType: str = '', state: str = '', beginId: str = '',
                           endId: str = '', begin: str = '', end: str = '', limit: str = '', proxies={},
                           proxy_host: str = None):
        '''
        获取最近7天挂单，且完全成交的订单数据，包括7天以前挂单，但近7天才成交的订单数据。按照订单创建时间倒序排序。
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-orders-last-7-days
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	否       	spread ID，如BTC-USDT_BTC-USDT-SWAP
        ordType           	String  	否       	订单类型limit：限价单
        state             	String  	否       	订单状态canceled：撤单成功filled：完全成交
        beginId           	String  	否       	请求的起始订单ID，请求此ID之后（更新的数据）的分页内容，不包括 beginId
        endId             	String  	否       	请求的结束订单ID，请求此ID之前（更旧的数据）的分页内容，不包括 endId
        begin             	String  	否       	筛选的开始时间戳，Unix 时间戳为毫秒数格式，如1597026383085
        end               	String  	否       	筛选的结束时间戳，Unix 时间戳为毫秒数格式，如1597027383085
        limit             	String  	否       	返回结果的数量，最大为100，默认100条
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	spread ID，如BTC-USDT_BTC-USDT-SWAP
        ordId             	String  	订单ID
        clOrdId           	String  	客户自定义订单ID
        tag               	String  	订单标签
        px                	String  	委托价格
        sz                	String  	委托数量
        ordType           	String  	订单类型limit：限价单
        side              	String  	订单方向
        fillSz            	String  	最新成交数量
        fillPx            	String  	最新成交价格
        tradeId           	String  	最近成交ID
        accFillSz         	String  	累计成交数量
        pendingFillSz     	String  	待成交数量（包括带结算数量）
        pendingSettleSz   	String  	待结算数量
        canceledSz        	String  	被取消数量
        avgPx             	String  	成交均价，如果成交数量为0，该字段为""
        state             	String  	订单状态canceled：撤单成功filled：完全成交
        cancelSource      	String  	撤单原因0: 系统撤单1: 用户撤单31: 当前只挂单订单 (Post only) 将会吃掉挂单深度32: 自成交保护撤单
        uTime             	String  	订单状态更新时间，Unix时间戳的毫秒数格式，如：1597026383085
        cTime             	String  	订单创建时间，Unix时间戳的毫秒数格式， 如 ：1597026383085
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_orders_history, **to_local(locals()))

    # 获取历史成交数据（近七天）
    def get_trades(self, sprdId: str = '', tradeId: str = '', ordId: str = '', beginId: str = '', endId: str = '',
                   begin: str = '', end: str = '', limit: str = '', proxies={}, proxy_host: str = None):
        '''
        获取近7天的订单成交明细信息. 结果按时间倒序返回。
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-trades-last-7-days
        
        限速：20次/2s
        限速规则：UserID
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	否       	spread ID，如BTC-USDT_BTC-USDT-SWAP
        tradeId           	String  	否       	交易 ID
        ordId             	String  	否       	订单 ID
        beginId           	String  	否       	请求的起始交易ID，请求此ID之后（更新的数据）的分页内容，不包括 beginId
        endId             	String  	否       	请求的结束交易ID，请求此ID之前（更旧的数据）的分页内容，不包括 endId
        begin             	String  	否       	筛选的开始时间戳，Unix 时间戳为毫秒数格式，如1597026383085
        end               	String  	否       	筛选的结束时间戳，Unix 时间戳为毫秒数格式，如1597027383085
        limit             	String  	否       	返回结果的数量，最大为100，默认100条
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	spread ID，如BTC-USDT_BTC-USDT-SWAP
        tradeId           	String  	交易ID
        ordId             	String  	订单ID
        clOrdId           	String  	客户自定义订单ID
        tag               	String  	订单标签
        fillPx            	String  	成交价格
        fillSz            	String  	成交数量
        side              	String  	交易方向buy：买sell：卖
        state             	String  	交易状态filled：已成交rejected：被拒绝
        execType          	String  	流动性方向T：takerM：maker
        ts                	String  	成交明细产生时间，Unix时间戳的毫秒数格式，如1597026383085
        legs              	Array of objects	交易的腿
        > instId          	String  	产品 ID
        > px              	String  	价格
        > sz              	String  	数量
        > side            	String  	交易方向buy：买sell：卖
        > fee             	String  	手续费金额或者返佣金额，手续费扣除为‘负数’，如-0.01；手续费返佣为‘正数’，如 0.01
        > feeCcy          	String  	交易手续费币种或者返佣金币种
        > tradeId         	String  	交易ID
        code              	String  	错误码，默认0
        msg               	String  	错误提示，默认 ""
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_trades, **to_local(locals()))

    # 获取Spreads（公共）
    def get_spreads(self, baseCcy: object = '', instId: str = '', sprdId: str = '', state: object = '', proxies={},
                    proxy_host: str = None):
        '''
        获取可交易的Spreads。
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-spreads-public
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        baseCcy           	string  	否       	Spread币种。 例如BTC, ETH
        instId            	String  	否       	Spread里包含的产品ID。
        sprdId            	String  	否       	Spread ID
        state             	string  	否       	Spread状态live：交易中suspend：暂停中expired：订单过期
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	spread ID
        sprdType          	String  	Spread类型，有效值为linear,inverse
        state             	String  	Spread状态live：交易中suspend：暂停中expired：已过期
        baseCcy           	String  	Spread币种。 例如BTC, ETH。
        szCcy             	String  	Spread数量单位。 例如USD, BTC, ETH, USD。
        quoteCcy          	String  	Spread计价单位。例如USDT，USD。
        tickSz            	String  	下单价格精度，如 0.0001。单位为Spread计价单位quoteCcy。
        minSz             	String  	最小下单数量。单位为Spread数量单位szCcy。
        lotSz             	String  	下单数量精度。单位为Spread数量单位szCcy。
        listTime          	String  	上线日期。Unix时间戳的毫秒数格式，如1597026383085
        expTime           	String  	失效日期。Unix时间戳的毫秒数格式，如1597026383085
        uTime             	String  	上次更新时间。Unix时间戳的毫秒数格式，如1597026383085
        legs              	array of objects	腿
        > instId          	String  	产品ID
        > side            	String  	产品方向。buy：买入sell：卖出
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_spreads, **to_local(locals()))

    # 获取Spread产品深度（公共）
    def get_books(self, sprdId: str, sz: str = '', proxies={}, proxy_host: str = None):
        '''
        获取Spread产品深度列表
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-order-book-public
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	是       	spread ID，如BTC-USDT_BTC-USDT-SWAP
        sz                	String  	否       	深度档位数量。最大值为400。默认值为5。
        返回参数:
        Parameter         	Type    	Description
        asks              	Array   	卖方深度
        bids              	Array   	买方深度
        ts                	String  	深度产生的时间
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_books, **to_local(locals()))

    # 获取单个Spread产品行情信息（公共）
    def get_ticker(self, sprdId: str, proxies={}, proxy_host: str = None):
        '''
        获取单个Spread产品行情信息，包括最新成交价，买一卖一价及数量。
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-ticker-public
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	是       	spread ID, e.g. BTC-USDT_BTC-USDT-SWAP
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	spread ID
        last              	String  	最新成交价
        lastSz            	String  	最新成交的数量
        askPx             	String  	卖一价
        askSz             	String  	卖一价对应的数量
        bidPx             	String  	买一价
        bidSz             	String  	买一价对应的数量
        ts                	String  	ticker数据产生时间，Unix时间戳的毫秒数格式，如1597026383085
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_ticker, **to_local(locals()))

    # 获取公共成交数据（公共）
    def get_public_trades(self, sprdId: str = '', proxies={}, proxy_host: str = None):
        '''
        查询市场上的Spread成交信息数据，每次请求最多返回500条结果。结果按时间倒序返回。
        https://www.okx.com/docs-v5/zh/#spread-trading-rest-api-get-public-trades-public
        
        限速：20次/2s
        限速规则：IP
    
        请求参数:
        Parameter         	Type    	Required	Description

        sprdId            	String  	否       	Spread ID，例如BTC-USDT_BTC-USDT-SWAP
        返回参数:
        Parameter         	Type    	Description
        sprdId            	String  	spread ID
        tradeId           	String  	交易ID
        px                	String  	成交价格
        sz                	String  	成交数量
        side              	String  	Taker的交易方向buy：买sell：卖
        ts                	String  	交易时间，Unix时间戳的毫秒数格式， 如 ：1597026383085
        '''
        return self.send_request(*_SpreadTradeEndpoints.get_public_trades, **to_local(locals()))
